# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import numpy as np
import abc
import enum
import logging
import time
import os
import math

# for SLEEP_TIME
import mpisppy.cylinders as cylinders
import mpisppy.utils.sputils as sputils

from mpi4py import MPI
from mpisppy.cylinders.spcommunicator import SPCommunicator


class ConvergerSpokeType(enum.Enum):
    OUTER_BOUND = 1
    INNER_BOUND = 2
    W_GETTER = 3
    NONANT_GETTER = 4

class Spoke(SPCommunicator):
    def __init__(self, spbase_object, fullcomm, strata_comm, cylinder_comm, options=None):
        super().__init__(spbase_object, fullcomm, strata_comm, cylinder_comm, options=None)
        self.local_write_id = 0
        self.remote_write_id = 0
        self.local_length = 0  # Does NOT include the + 1
        self.remote_length = 0  # Length on hub; does NOT include + 1

        self.last_call_to_got_kill_signal = time.time()

    def _make_windows(self, local_length, remote_length):
        # Spokes notify the hub of the buffer sizes
        pair_of_lengths = np.array([local_length, remote_length], dtype="i")
        self.strata_comm.Send((pair_of_lengths, MPI.INT), dest=0, tag=self.strata_rank)
        self.local_length = local_length
        self.remote_length = remote_length
        
        # Make the windows of the appropriate buffer sizes
        # To do?: Spoke should not need to know how many other spokes there are.
        # Just call a single _make_window()? Do you need to create empty
        # windows?
        # ANSWER (dlw July 2020): Since the windows have zero length and since
        # the number of spokes is not expected to be large, it is probably OK.
        # The (minor) benefit is that free_windows does not need to know if it
        # was called by a hub or a spoke. If we ever move to dynamic spoke
        # creation, then this needs to be reimagined.
        self.windows = [None for _ in range(self.n_spokes)]
        self.buffers = [None for _ in range(self.n_spokes)]
        for i in range(self.n_spokes):
            length = self.local_length if self.strata_rank == i + 1 else 0
            win, buff = self._make_window(length)
            self.windows[i] = win
            self.buffers[i] = buff

        self._windows_constructed = True

    def spoke_to_hub(self, values):
        """ Put the specified values into the locally-owned buffer for the hub
            to pick up.

            Notes:
                This automatically does the -1 indexing

                This assumes that values contains a slot at the end for the
                write_id
        """
        expected_length = self.local_length + 1
        if len(values) != expected_length:
            raise RuntimeError(
                f"Attempting to put array of length {len(values)} "
                f"into local buffer of length {expected_length}"
            )
        self.local_write_id += 1
        values[-1] = self.local_write_id
        window = self.windows[self.strata_rank - 1]
        window.Lock(self.strata_rank)
        window.Put((values, len(values), MPI.DOUBLE), self.strata_rank)
        window.Unlock(self.strata_rank)

    def spoke_from_hub(self, values):
        """
        """
        expected_length = self.remote_length + 1
        if len(values) != expected_length:
            raise RuntimeError(
                f"Spoke trying to get buffer of length {expected_length} "
                f"from hub, but provided buffer has length {len(values)}."
            )
        window = self.windows[self.strata_rank - 1]
        window.Lock(0)
        window.Get((values, len(values), MPI.DOUBLE), 0)
        window.Unlock(0)

        if values[-1] > self.remote_write_id:
            self.remote_write_id = values[-1]
            return True
        return False

    def got_kill_signal(self):
        """ Spoke should call this method at least every iteration
            to see if the Hub terminated
        """
        # Spokes can sometimes call this frequently in a tight loop,
        # causing the Allreduces to become out of sync
        diff = time.time() - self.last_call_to_got_kill_signal
        if diff < cylinders.SPOKE_SLEEP_TIME:
            time.sleep(cylinders.SPOKE_SLEEP_TIME - diff)
        self.last_call_to_got_kill_signal = time.time()
        return self._got_kill_signal()

    @abc.abstractmethod
    def main(self):
        """
        The main call for the Spoke. Derived classe
        should call the got_kill_signal method 
        regularly to ensure all ranks terminate 
        with the Hub.
        """
        pass

    @abc.abstractmethod
    def get_serial_number(self):
        pass

    @abc.abstractmethod
    def _got_kill_signal(self):
        """ Every spoke needs a way to get the signal to terminate
            from the hub
        """
        pass


class _BoundSpoke(Spoke):
    """ A base class for bound spokes
    """
    def __init__(self, spbase_object, fullcomm, strata_comm, cylinder_comm, options=None):
        super().__init__(spbase_object, fullcomm, strata_comm, cylinder_comm, options=None)
        if self.cylinder_rank == 0 and \
                'trace_prefix' in spbase_object.options and \
                spbase_object.options['trace_prefix'] is not None:
            trace_prefix = spbase_object.options['trace_prefix']

            filen = trace_prefix+self.__class__.__name__+'.csv'
            if os.path.exists(filen):
                raise RuntimeError(f"Spoke trace file {filen} already exists!")
            with open(filen, 'w') as f:
                f.write("time,bound\n")
            self.trace_filen = filen
            self.start_time = spbase_object.start_time
        else:
            self.trace_filen = None

    def make_windows(self):
        """ Makes the bound window and a remote window to
            look for a kill signal
        """

        ## need a remote_length for the kill signal
        self._make_windows(1, 0)
        self._kill_sig = np.zeros(0 + 1)
        self._bound = np.zeros(1 + 1)

    @property
    def bound(self):
        return self._bound[0]

    @bound.setter
    def bound(self, value):
        self._append_trace(value)
        self._bound[0] = value
        self.spoke_to_hub(self._bound)

    def get_serial_number(self):
        return int(self._kill_sig[-1])

    def _got_kill_signal(self):
        """Looks for the kill signal and returns True if sent"""
        self.spoke_from_hub(self._kill_sig)
        kill = self._kill_sig[-1] == -1
        return kill

    def _append_trace(self, value):
        if self.cylinder_rank != 0 or self.trace_filen is None:
            return
        with open(self.trace_filen, 'a') as f:
            f.write(f"{time.perf_counter()-self.start_time},{value}\n")


class _BoundNonantLenSpoke(_BoundSpoke):
    """ A base class for bound spokes which also
        want something of len nonants from OPT
    """

    def make_windows(self):
        """ Makes the bound window and with a remote buffer long enough
            to hold an array as long as the nonants.

            Input:
                opt (SPBase): Must have local_scenarios attached already!

        """
        if not hasattr(self.opt, "local_scenarios"):
            raise RuntimeError("Provided SPBase object does not have local_scenarios attribute")

        if len(self.opt.local_scenarios) == 0:
            raise RuntimeError(f"Rank has zero local_scenarios")

        vbuflen = 0
        for s in self.opt.local_scenarios.values():
            vbuflen += len(s._mpisppy_data.nonant_indices)

        self._make_windows(1, vbuflen)
        self._locals = np.zeros(vbuflen + 1) # Also has kill signal
        self._bound = np.zeros(1 + 1)
        self._new_locals = False

    def get_serial_number(self):
        return int(self._locals[-1])

    def _got_kill_signal(self):
        """ returns True if a kill signal was received, 
            and refreshes the array and _locals"""
        self._new_locals = self.spoke_from_hub(self._locals)
        kill = self._locals[-1] == -1
        return kill


class InnerBoundSpoke(_BoundSpoke):
    """ For Spokes that provide an inner bound through self.bound to the
        Hub, and do not need information from the main PH OPT hub.
    """
    converger_spoke_types = (ConvergerSpokeType.INNER_BOUND,)
    converger_spoke_char = 'I'


class OuterBoundSpoke(_BoundSpoke):
    """ For Spokes that provide an outer bound through self.bound to the
        Hub, and do not need information from the main PH OPT hub.
    """
    converger_spoke_types = (ConvergerSpokeType.OUTER_BOUND,)
    converger_spoke_char = 'O'


class _BoundWSpoke(_BoundNonantLenSpoke):
    """ A base class for bound spokes which also want the W's from the OPT
        threads
    """

    @property
    def localWs(self):
        """Returns the local copy of the weights"""
        return self._locals[:-1]

    @property
    def new_Ws(self):
        """ Returns True if the local copy of 
            the weights has been updated since
            the last call to got_kill_signal
        """
        return self._new_locals


class OuterBoundWSpoke(_BoundWSpoke):
    """
    For Spokes that provide an outer bound
    through self.bound to the Hub,
    and receive the Ws (or weights) from
    the main PH OPT hub.
    """

    converger_spoke_types = (
        ConvergerSpokeType.OUTER_BOUND,
        ConvergerSpokeType.W_GETTER,
    )
    converger_spoke_char = 'O'


class _BoundNonantSpoke(_BoundNonantLenSpoke):
    """ A base class for bound spokes which also
        want the xhat's from the OPT threads
    """

    @property
    def localnonants(self):
        """Returns the local copy of the nonants"""
        return self._locals[:-1]

    @property
    def new_nonants(self):
        """Returns True if the local copy of 
           the nonants has been updated since
           the last call to got_kill_signal"""
        return self._new_locals


class InnerBoundNonantSpoke(_BoundNonantSpoke):
    """ For Spokes that provide an inner (incumbent) 
        bound through self.bound to the Hub,
        and receive the nonants from
        the main SPOpt hub.

        Includes some helpful methods for saving
        and restoring results
    """
    converger_spoke_types = (
        ConvergerSpokeType.INNER_BOUND,
        ConvergerSpokeType.NONANT_GETTER,
    )
    converger_spoke_char = 'I'

    def __init__(self, spbase_object, fullcomm, strata_comm, cylinder_comm, options=None):
        super().__init__(spbase_object, fullcomm, strata_comm, cylinder_comm, options=None)
        self.is_minimizing = self.opt.is_minimizing
        self.best_inner_bound = math.inf if self.is_minimizing else -math.inf
        self.solver_options = None # can be overwritten by derived classes

        # NOTE: defaults to True
        self.save_tree_solution = True
        if ('save_tree_solution' in self.options) and (not self.options['save_tree_solution']):
            self.save_tree_solution = False

        # set up best nonant cache
        # NOTE: should we also cache the tree solution??
        for k,s in self.opt.local_scenarios.items():
            s._mpisppy_data.best_nonant_cache = None

    def update_if_improving(self, candidate_inner_bound):
        if candidate_inner_bound is None:
            return False
        update = (candidate_inner_bound < self.best_inner_bound) \
                if self.is_minimizing else \
                (self.best_inner_bound < candidate_inner_bound)
        if not update:
            return False

        self.best_inner_bound = candidate_inner_bound
        # send to hub
        self.bound = candidate_inner_bound
        self._cache_best_nonants()
        return True

    def finalize(self):
        self._restore_and_fix_best_nonants()
        if not self.opt.first_stage_solution_available:
            return None
        if not self.save_tree_solution:
            return None

        self.opt.solve_loop(solver_options=self.solver_options,
                           verbose=False,
                           tee=False)

        ## NOTE: this should be feasible here,
        ##       if not we've done something wrong
        feasP = self.opt.feas_prob()
        if abs(feasP - self.opt.E1) > self.opt.E1_tolerance:
            raise RuntimeError(f"Found infeasible solution which was feasible before - feasP={feasP}, E1={self.opt.E1}, E1_tolerance={self.opt.E1_tolerance}")

        obj = self.opt.Eobjective(verbose=False)

        if not math.isclose(obj,self.best_inner_bound):
            if self.cylinder_rank == 0:
                print(f"WARNING: {self.__class__.__name__} best inner bound is different "
                        f"from objective calculated in finalize")
                print(f"Best inner bound: {self.best_inner_bound}")
                print(f"Current objective: {obj}")

        self.opt.tree_solution_available = True
        self.final_bound = obj
        return obj

    def _cache_best_nonants(self):
        for k,s in self.opt.local_scenarios.items():
            scenario_cache = {}
            for ndn_i, var in s._mpisppy_data.nonant_indices.items():
                scenario_cache[ndn_i] = var.value
            s._mpisppy_data.best_nonant_cache = scenario_cache

    def _restore_and_fix_best_nonants(self):
        for k,s in self.opt.local_scenarios.items():
            scenario_cache = s._mpisppy_data.best_nonant_cache
            if scenario_cache is None:
                return
            is_persistent = sputils.is_persistent(s._solver_plugin)
            solver = s._solver_plugin
            for ndn_i, var in s._mpisppy_data.nonant_indices.items():
                var.fix(scenario_cache[ndn_i])
                if is_persistent:
                    solver.update_var(var)
        self.opt.first_stage_solution_available = True


class OuterBoundNonantSpoke(_BoundNonantSpoke):
    """ For Spokes that provide an outer
        bound through self.bound to the Hub,
        and receive the nonants from
        the main OPT hub.
    """
    converger_spoke_types = (
        ConvergerSpokeType.OUTER_BOUND,
        ConvergerSpokeType.NONANT_GETTER,
    )
    converger_spoke_char = 'A'  # probably Lagrangian



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# Provide some test for aph under mpi-sppy.
# Author: David L. Woodruff (started circa September 2019)
#         add farmer tests April 2021
"""
IMPORTANT:
  Unless we run to convergence and check xhat, the solver, and even solver
version matter a lot, so we often just do smoke tests for sizes.
"""

import unittest
from math import log10, floor
import pyomo.environ as pyo
import mpisppy.opt.aph
import mpisppy.phbase
from mpisppy.tests.examples.sizes.sizes import scenario_creator, \
                                               scenario_denouement, \
                                               _rho_setter
import mpisppy.tests.examples.farmer as farmer
from mpisppy.tests.test_utils import get_solver, round_pos_sig

__version__ = 0.6
solver_available,solvername, persistent_available, persistentsolvername= get_solver()

import mpi4py.MPI as mpi
fullcomm = mpi.COMM_WORLD
global_rank = fullcomm.Get_rank()

#*****************************************************************************
class Test_aph_sizes(unittest.TestCase):
    """ Test the aph mpisppy code using sizes."""

    def setUp(self):
        self.Baseoptions = {}
        self.Baseoptions["asynchronousPH"] = False
        self.Baseoptions["solvername"] = solvername
        self.Baseoptions["PHIterLimit"] = 10
        self.Baseoptions["defaultPHrho"] = 1
        self.Baseoptions["convthresh"] = 0.001
        self.Baseoptions["subsolvedirectives"] = None
        self.Baseoptions["verbose"] = False
        self.Baseoptions["display_timing"] = False
        self.Baseoptions["display_progress"] = False
        self.Baseoptions["iter0_solver_options"] = {"mipgap": 0.1}
        self.Baseoptions["iterk_solver_options"] = {"mipgap": 0.02}

        self.Baseoptions["display_progress"] = False

        self.all3_scenario_names = list()
        for sn in range(3):
            self.all3_scenario_names.append("Scenario"+str(sn+1))

        self.all10_scenario_names = list()
        for sn in range(10):
            self.all10_scenario_names.append("Scenario"+str(sn+1))

    def _copy_of_base_options(self):
        retval = {}
        for k,v in self.Baseoptions.items():
            retval[k] = v
        return retval
        
    def test_make_aph(self):
        """ Just smoke to verify construction"""
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["async_frac_needed"] = 0.5
        options["async_sleep_secs"] = 0.5
        aph = mpisppy.opt.aph.APH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )

    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_aph_basic(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["async_frac_needed"] = 0.5
        options["async_sleep_secs"] = 0.5
        aph = mpisppy.opt.aph.APH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )

        conv, obj, tbound = aph.APH_main(spcomm=None)
        print ("objthing={}, (was=-2435908)".format(obj))
        print ("tbound ={} (was=224106)".format(tbound))

    def test_bundles(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["bundles_per_rank"] = 2
        options["async_frac_needed"] = 0.5
        options["async_sleep_secs"] = 0.5
        aph = mpisppy.opt.aph.APH(
            options,
            self.all10_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 10},
        )
        conv, obj, tbound = aph.APH_main(spcomm=None)
        print ("bundle objthing={}, (was=224712.9)".format(obj))
        print ("bundle tbound ={} (was=223168.5)".format(tbound))

        
    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_APHgamma(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["async_frac_needed"] = 0.5
        options["async_sleep_secs"] = 0.5
        a = 2
        options["APHgamma"] = a
        aph = mpisppy.opt.aph.APH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )

        conv, obj, tbound = aph.APH_main(spcomm=None)
        print (f"APHgamma={a}; objthing={obj}")
        print ("tbound ={}".format(tbound))


    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_use_lag(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["async_frac_needed"] = 0.5
        options["async_sleep_secs"] = 0.5
        a = 2
        options["APHuse_lag"] = True
        aph = mpisppy.opt.aph.APH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )

        conv, obj, tbound = aph.APH_main(spcomm=None)
        print (f"use lag objthing={obj}")
        print ("tbound ={}".format(tbound))


    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_running_dump(self):
        # just see if "display_convergence_detail" causes a crash
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["display_convergence_detail"] = True
        options["async_frac_needed"] = 1
        options["async_sleep_secs"] = 0.5
        aph = mpisppy.opt.aph.APH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
        conv, obj, tbound = aph.APH_main(spcomm=None)


    def test_lags_bundles(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["bundles_per_rank"] = 2
        options["async_frac_needed"] = 0.5
        options["async_sleep_secs"] = 0.5
        options["APHuse_lag"] = True
        aph = mpisppy.opt.aph.APH(
            options,
            self.all10_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 10},
        )
        conv, obj, tbound = aph.APH_main(spcomm=None)
        print ("uselag bundle objthing={}, (pre-lag was=224712.9)".format(obj))
        print ("bundle tbound ={} (was=223168.5)".format(tbound))

#*****************************************************************************
class Test_aph_farmer(unittest.TestCase):
    """ Test the aph mpisppy code using farmer."""

    def setUp(self):
        self.Baseoptions = {}
        self.Baseoptions["asynchronousPH"] = False
        self.Baseoptions["solvername"] = solvername
        self.Baseoptions["PHIterLimit"] = 10
        self.Baseoptions["defaultPHrho"] = 1
        self.Baseoptions["convthresh"] = 0.001
        self.Baseoptions["subsolvedirectives"] = None
        self.Baseoptions["verbose"] = False
        self.Baseoptions["display_timing"] = False
        self.Baseoptions["display_progress"] = False
        self.Baseoptions["iter0_solver_options"] = None
        self.Baseoptions["iterk_solver_options"] = None

        self.Baseoptions["display_progress"] = False

    def _copy_of_base_options(self):
        retval = {}
        for k,v in self.Baseoptions.items():
            retval[k] = v
        return retval

    def round_pos_sig(self, x, sig=1):
        return round(x, sig-int(floor(log10(abs(x))))-1)

    def _make_scenario_names(self, cnt):
        return [f"Scenario{i+1}" for i in range(cnt)]
        
    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_aph_farmer_basic30(self):
        Aoptions = self._copy_of_base_options()
        Aoptions["PHIterLimit"] = 2
        Aoptions["aph_dispatch_frac"] = 1
        Aoptions["async_frac_needed"] = 1
        Aoptions["async_sleep_secs"] = 0.01
        Aoptions["aph_gamma"] = 1
        aph = mpisppy.opt.aph.APH(
            Aoptions,
            self._make_scenario_names(30),
            farmer.scenario_creator,
            farmer.scenario_denouement,
            scenario_creator_kwargs={"crops_multiplier": 1},
        )

        conv, obj, tbound = aph.APH_main(spcomm=None)
        digits=3
        tbtarget = self.round_pos_sig(137846, digits)
        boundgot = self.round_pos_sig(-tbound, digits)
        self.assertEqual(tbtarget, boundgot)
        # Note: this is not a valid xhat obj, it is weighted
        objtarget = self.round_pos_sig(134796, digits)
        objgot = self.round_pos_sig(-obj, digits)
        self.assertEqual(objgot, objtarget)

    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_aph_farmer_dispatch(self):
        Aoptions = self._copy_of_base_options()
        Aoptions["PHIterLimit"] = 2
        Aoptions["dispatch_frac"] = 0.25
        Aoptions["async_frac_needed"] = 1
        Aoptions["async_sleep_secs"] = 0.01
        Aoptions["aph_gamma"] = 1
        aph = mpisppy.opt.aph.APH(
            Aoptions,
            self._make_scenario_names(30),
            farmer.scenario_creator,
            farmer.scenario_denouement,
            scenario_creator_kwargs={"crops_multiplier": 1},
        )

        conv, obj, tbound = aph.APH_main(spcomm=None)
        digits=3
        tbtarget = self.round_pos_sig(137846, digits)
        boundgot = self.round_pos_sig(-tbound, digits)
        self.assertEqual(tbtarget, boundgot)
        # Note: this is not a valid xhat obj, it is weighted
        objtarget = self.round_pos_sig(134680, digits)
        objgot = self.round_pos_sig(-obj, digits)
        self.assertEqual(objgot, objtarget)


    @unittest.skipIf(not solver_available,
                     "%s solver is not available" % (solvername,))
    def test_aph_farmer_dispatch_bundles(self):
        Aoptions = self._copy_of_base_options()
        Aoptions["PHIterLimit"] = 2
        Aoptions["dispatch_frac"] = 0.25
        Aoptions["async_frac_needed"] = 1
        Aoptions["async_sleep_secs"] = 0.01
        Aoptions["aph_gamma"] = 1
        Aoptions["bundles_per_rank"] = 5
        
        aph = mpisppy.opt.aph.APH(
            Aoptions,
            self._make_scenario_names(30),
            farmer.scenario_creator,
            farmer.scenario_denouement,
            scenario_creator_kwargs={"crops_multiplier": 1},
        )

        conv, obj, tbound = aph.APH_main(spcomm=None)
        digits=3
        tbtarget = self.round_pos_sig(133005, digits)
        boundgot = self.round_pos_sig(-tbound, digits)
        self.assertEqual(tbtarget, boundgot)
        # Note: this is not a valid xhat obj, it is weighted
        objtarget = self.round_pos_sig(132959, digits)
        objgot = self.round_pos_sig(-obj, digits)
        self.assertEqual(objgot, objtarget)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import os
import mpisppy.tests.examples.sizes.ReferenceModel as ref
import mpisppy.utils.sputils as sputils

def scenario_creator(scenario_name, scenario_count=None):
    """ The callback needs to create an instance and then attach
        the PySP nodes to it in a list _mpisppy_node_list ordered by stages.
        Optionally attach _PHrho. 
    """
    if scenario_count is None:
        raise ValueError("Sizes scenario_creator requires a scenario_count kwarg")
    if scenario_count not in [3, 10]:
        raise ValueError("Sizes scenario count must equal either 3 or 10")

    sizes_dir = os.path.dirname(__file__)
    datadir = os.sep.join((sizes_dir, f"SIZES{scenario_count}"))
    try:
        fname = datadir + os.sep + scenario_name + ".dat"
    except:
        print("FAIL: datadir=", datadir, " scenario_name=", scenario_name)

    model = ref.model.create_instance(fname)

    # now attach the one and only tree node
    varlist = [model.NumProducedFirstStage, model.NumUnitsCutFirstStage]
    sputils.attach_root_node(model, model.FirstStageCost, varlist)

    return model


def scenario_denouement(rank, scenario_name, scenario):
    pass


def _rho_setter(scen):
    """ rho values for the scenario.
    Args:
        scen (pyo.ConcreteModel): the scenario
    Returns:
        a list of (id(vardata), rho)
    """
    retlist = []
    RF = 0.001  # a factor for rho, if you like
    cutrho = scen.UnitReductionCost * RF

    for i in scen.ProductSizes:
        idv = id(scen.NumProducedFirstStage[i])
        rho = scen.UnitProductionCosts[i] * RF
        retlist.append((idv, rho))

        for j in scen.ProductSizes:
            if j <= i:
                idv = id(scen.NumUnitsCutFirstStage[i, j])
                retlist.append((idv, cutrho))

    return retlist


def id_fix_list_fct(s):
    """ specify tuples used by the fixer.

    Args:
        s (ConcreteModel): the sizes instance.
    Returns:
         i0, ik (tuples): one for iter 0 and other for general iterations.
             Var id,  threshold, nb, lb, ub
             The threshold is on the square root of the xbar squared differnce
             nb, lb an bu an "no bound", "upper" and "lower" and give the numver
                 of iterations or None for ik and for i0 anything other than None
                 or None. In both cases, None indicates don't fix.
    """
    import mpisppy.extensions.fixer as fixer

    iter0tuples = []
    iterktuples = []
    for i in s.ProductSizes:
        iter0tuples.append(
            fixer.Fixer_tuple(s.NumProducedFirstStage[i], th=0.01, nb=None, lb=0, ub=0)
        )
        iterktuples.append(
            fixer.Fixer_tuple(s.NumProducedFirstStage[i], th=0.2, nb=3, lb=1, ub=2)
        )
        for j in s.ProductSizes:
            if j <= i:
                iter0tuples.append(
                    fixer.Fixer_tuple(
                        s.NumUnitsCutFirstStage[i, j], th=0.5, nb=None, lb=0, ub=0
                    )
                )
                iterktuples.append(
                    fixer.Fixer_tuple(
                        s.NumUnitsCutFirstStage[i, j], th=0.2, nb=3, lb=1, ub=2
                    )
                )

    return iter0tuples, iterktuples



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' Utilities for reading and writing W and 
    x-bar values in and out of csv files. 

    Written: DLW July 2019
    Modified: DTM Aug 2019

    Could stand to be re-factored a bit.

    When reading/writing the weight files, there are two options.
        
     1. All weights are stored in a single main file. 
        
            Reading: Set "init_W_fname" to the location of the main file
            containing all of the weights.

            Writing: Set "W_fname" to the location of the main file that will
            contain all the weights.

     2. Each scenario's weights are stored in separate, individual files.
        
            Reading: Set "init_W_fname" to the directory containing the weight
            files. This directory must contain one file per scenario, each
            named <scenario_name>_weights.csv. Set "init_separate_W_files" to
            True in the options dictionary.

            Writing: Set "W_fname" to the directory that will contain the
            weight files (this directory will be created if it does not already
            exist). The weight files will be corretly named in accordance with
            the naming convention above. Set "separate_W_files" to True in the
            options dictionary.
'''

import pyomo.environ as pyo
import os

''' W utilities '''

def write_W_to_file(PHB, fname, sep_files=False):
    '''
    Args:
        PHB (PHBase object) -- Where the W values live
        fname (str) -- name of file to which we write.
        sep_files (bool, optional) -- If True, one file will be written for
            each scenario, rather than one main file. The names of the files
            are the names of the scenarios.

    Notes:
        All ranks pass their information to rank 0, which then writes a single
        file. This can apparently be accomplished using Collective MPI I/O (see
        https://mpi4py.readthedocs.io/en/stable/tutorial.html#mpi-io), but I am
        lazy and doing this for now.
    '''

    if (sep_files):
        for (sname, scenario) in PHB.local_scenarios.items():
            scenario_Ws = {var.name: pyo.value(scenario._mpisppy_model.W[node.name, ix])
                for node in scenario._mpisppy_node_list
                for (ix, var) in enumerate(node.nonant_vardata_list)}
            scenario_fname = os.path.join(fname, sname + '_weights.csv')
            with open(scenario_fname, 'w') as f:
                for (vname, val) in scenario_Ws.items():
                    row = ','.join([vname, str(val)]) + '\n'
                    f.write(row)
    else:
        local_Ws = {(sname, var.name): pyo.value(scenario._mpisppy_model.W[node.name, ix])
                    for (sname, scenario) in PHB.local_scenarios.items()
                    for node in scenario._mpisppy_node_list
                    for (ix, var) in enumerate(node.nonant_vardata_list)}
        comm = PHB.comms['ROOT']
        Ws = comm.gather(local_Ws, root=0)
        if (PHB.cylinder_rank == 0):
            with open(fname, 'a') as f:
                for W in Ws:
                    for (key, val) in W.items():
                        sname, vname = key[0], key[1]
                        row = ','.join([sname, vname, str(val)]) + '\n'
                        f.write(row)

def set_W_from_file(fname, PHB, rank, sep_files=False):
    ''' 
    Args:
        fname (str) -- if sep_files=False, file containing the dual weights.
            Otherwise, path of the directory containing the dual weight files
            (one per scenario).
        PHB (PHBase object) -- Where the W values will be put
        rank (int) -- rank number
        sep_files (bool, optional) -- If True, attempt to read weights from
            individual files, one per scenario. The files must be contained in
            the same directory, and must be named <sname>_weights.csv for each
            scenario name <sname>.
    
    Notes:
        Calls _check_W, which ensures that all required values were specified,
        and that the specified weights satisfy the dual feasibility condition 
        sum_{s\in S} p_s * w_s = 0.
    '''
    scenario_names_local  = list(PHB.local_scenarios.keys())
    scenario_names_global = PHB.all_scenario_names

    if (sep_files):
        w_val_dict = dict()
        for sname in scenario_names_local:
            scenario_fname = os.path.join(fname, sname + '_weights.csv')
            w_val_dict[sname] = _parse_W_csv_single(scenario_fname)
    else:
        w_val_dict = _parse_W_csv(fname, scenario_names_local,
                                    scenario_names_global, rank)

    _check_W(w_val_dict, PHB, rank)

    mp = {(sname, var.name): (node.name, ix)
            for (sname, scenario) in PHB.local_scenarios.items()
            for node in scenario._mpisppy_node_list
            for (ix,var) in enumerate(node.nonant_vardata_list)}

    for (sname, d) in w_val_dict.items():
        for vname in d.keys():
            scenario = PHB.local_scenarios[sname]
            node_name, ix = mp[sname, vname]
            scenario._mpisppy_model.W[node_name, ix] = w_val_dict[sname][vname]

def _parse_W_csv_single(fname):
    ''' Read a file containing the weights for a single scenario. The file must
        be formatted as 

        variable_name,variable_value

        (comma separated). Lines beginning with a "#" are treated as comments
        and ignored.
    '''
    if (not os.path.exists(fname)):
        raise RuntimeError('Could not find file {fn}'.format(fn=fname))
    results = dict()
    with open(fname, 'r') as f:
        for line in f:
            if (line.startswith('#')):
                continue
            line  = line.split(',')
            vname = ','.join(line[:-1])
            wval  = float(line[-1])
            results[vname] = wval
    return results

def _parse_W_csv(fname, scenario_names_local, scenario_names_global, rank):
    ''' Read a csv file containing weight information. 
        
        Args:
            fname (str) -- Filename of csv file to read
            scenario_names_local (list of str) -- List of local scenario names
            scenario_names_global (list of str) -- List of global scenario
                names (i.e. all the scenario names in the entire model across
                all ranks--each PHBase object stores this information).

        Return:
            results (dict) -- Doubly-nested dict mapping 
                results[scenario_name][var_name] --> weight value (float)
    
        Notes:
            This function is only called if sep_files=False, i.e., if all of
            the weights are stored in a single root file. The file must be
            formatted as:

            scenario_name,variable_name,weight_value

            Rows that begin with a "#" character are treated as comments.
            The variable names _may_ contain commas (confusing, but simpler for
            the user)

            Raises a RuntimeError if there are any missing scenarios. Prints a
            warning if there are any extra scenarios.

            When this function returns, we are certain that 
                results.keys() == PHB.local_scenarios.keys()

            When run in parallel, this method requires multiple ranks to open
            and read from the same file simultaneously. Apparently there are
            safer ways to do this using MPI collective communication, but since
            all we're doing here is reading files, I'm being lazy and doing it
            this way.
    '''
    results = dict()
    seen = {name: False for name in scenario_names_local}
    with open(fname, 'r') as f:
        for line in f:
            if (line.startswith('#')):
                continue
            line  = line.split(',')
            sname = line[0]
            vname = ','.join(line[1:-1])
            wval  = float(line[-1])
            
            if (sname not in scenario_names_global):
                if (rank == 0):
                    print('WARNING: Ignoring unknown scenario name', sname)
                continue
            if (sname not in scenario_names_local):
                continue
            if (sname in results):
                results[sname][vname] = wval
            else:
                seen[sname] = True
                results[sname] = {vname: wval}
    missing = [name for (name,is_seen) in seen.items() if not is_seen]
    if (missing):
        raise RuntimeError('rank ' + str(rank) +' could not find the following '
                'scenarios in the provided weight file: ' + ', '.join(missing)) 
        
    return results
            
def _check_W(w_val_dict, PHB, rank):
    '''
    Args:
        w_val_dict (dict) -- doubly-nested dict mapping 
            w_val_dict[scenario_name][variable_name] = weight value.
        PHB (PHBase object) -- PHBase object
        rank (int) -- local rank

    Notes:
        Checks for three conditions:
         
         1. Missing variables --> raises a RuntimeError
         2. Extra variables --> prints a warning
         3. Dual feasibility --> raises a RuntimeError
    '''
    # By this point, we are certain that
    # w_val_dict.keys() == PHB.local_scenarios.keys()
    for (sname, scenario) in PHB.local_scenarios.items():
        vn_model = set([var.name for node in scenario._mpisppy_node_list
                                 for var  in node.nonant_vardata_list])
        vn_provided = set(w_val_dict[sname].keys())
        diff = vn_model.difference(vn_provided)
        if (diff):
            raise RuntimeError(sname + ' is missing '
                'the following variables: ' + ', '.join(list(diff)))
        diff = vn_provided.difference(vn_model)
        if (diff):
            print('Removing unknown variables:', ', '.join(list(diff)))
            for vname in diff:
                w_val_dict[sname].pop(vname, None)
        
    # At this point, we are sure that every local 
    # scenario has the same set of variables
    probs = {name: model._mpisppy_probability for (name, model) in
                    PHB.local_scenarios.items()}

    checks = dict()
    for vname in vn_model: # Ensured vn_model = vn_provided
        checks[vname] = sum(probs[name] * w_val_dict[name][vname]
                            for name in PHB.local_scenarios.keys())

    checks = PHB.comms['ROOT'].gather(checks, root=0)
    if (rank == 0):
        for vname in vn_model:
            dual = sum(c[vname] for c in checks)
            if (abs(dual) > 1e-7):
                raise RuntimeError('Provided weights do not satisfy '
                    'dual feasibility: \sum_{scenarios} prob(s) * w(s) != 0. '
                    'Error on variable ' + vname)

''' X-bar utilities '''

def write_xbar_to_file(PHB, fname):
    '''
    Args:
        PHB (PHBase object) -- Where the W values live
        fname (str) -- name of file to which we write.

    Notes:
        Each scenario maintains its own copy of xbars. We only need to write
        one of them to the file (i.e. no parallelism required).
    '''
    if (PHB.cylinder_rank != 0):
        return
    sname = list(PHB.local_scenarios.keys())[0]
    scenario = PHB.local_scenarios[sname]
    xbars = {var.name: pyo.value(scenario._mpisppy_model.xbars[node.name, ix])
                for node in scenario._mpisppy_node_list
                for (ix, var) in enumerate(node.nonant_vardata_list)}
    with open(fname, 'a') as f:
        for (var_name, val) in xbars.items():
            row = ','.join([var_name, str(val)]) + '\n'
            f.write(row)

def set_xbar_from_file(fname, PHB):
    ''' Read all of the csv files in a directory and use them to populate
        _xbars and _xsqbars

    Args:
        fname (str) -- file containing the dual weights
        PHB (PHBase object) -- Where the W values will be put

    Notes:
        Raises a RuntimeError if the provided file is missing any values for
        xbar (i.e. does not assume a default value for missing variables).
    '''
    xbar_val_dict = _parse_xbar_csv(fname)

    if (PHB.cylinder_rank == 0):
        _check_xbar(xbar_val_dict, PHB)

    for (sname, scenario) in PHB.local_scenarios.items():
        for node in scenario._mpisppy_node_list:
            for (ix,var) in enumerate(node.nonant_vardata_list):
                val = xbar_val_dict[var.name]
                scenario._mpisppy_model.xbars[node.name, ix] = val
                scenario._mpisppy_model.xsqbars[node.name, ix] = val * val

def _parse_xbar_csv(fname):
    ''' Read a csv file containing weight information. 
        
        Args:
            fname (str) -- Filename of csv file to read
        Return:
            results (dict) -- Dict mapping var_name --> variable value (float)
    
        Notes:
            The file must be formatted as:

            variable_name,value

            Rows that begin with a "#" character are treated as comments.
            The variable names _may_ contain commas (confusing, but simpler for
            the user)

            When run in parallel, this method requires multiple ranks to open
            and read from the same file simultaneously. Apparently there are
            safer ways to do this using MPI collective communication, but since
            all we're doing here is reading files, I'm being lazy and doing it
            this way.
    '''
    results = dict()
    with open(fname, 'r') as f:
        for line in f:
            if (line.startswith('#')):
                continue
            line  = line.split(',')
            vname = ','.join(line[:-1])
            val  = float(line[-1])
            
            results[vname] = val

    return results

def _check_xbar(xbar_val_dict, PHB):
    ''' Make sure that a value was provided for every non-anticipative
        variable. If any extra variable values were provided in the input file,
        this function prints a warning.
    '''
    sname = list(PHB.local_scenarios.keys())[0]
    scenario = PHB.local_scenarios[sname]
    var_names = set([var.name for node in scenario._mpisppy_node_list
                          for var  in node.nonant_vardata_list])
    provided_vars = set(xbar_val_dict.keys())
    set1 = var_names.difference(provided_vars)
    if (set1):
        raise RuntimeError('Could not find the following required variable '
            'values in the provided input file: ' + ', '.join([v for v in set1]))
    set2 = provided_vars.difference(var_names)
    if (set2):
        print('Ignoring the following variables values provided in the '
              'input file: ' + ', '.join([v for v in set2]))



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# base class for hub and for spoke strata
import logging
import time
import math
import inspect

import numpy as np
from mpi4py import MPI

import pyomo.environ as pyo
from pyomo.opt import SolverFactory, SolutionStatus, TerminationCondition

from mpisppy import global_toc
from mpisppy.spbase import SPBase
import mpisppy.utils.sputils as sputils

logger = logging.getLogger("SPOpt")
logger.setLevel(logging.WARN)

class SPOpt(SPBase):
    """ Defines optimization methods for hubs and spokes """

    def __init__(
            self,
            options,
            all_scenario_names,
            scenario_creator,
            scenario_denouement=None,
            all_nodenames=None,
            mpicomm=None,
            extensions=None,
            extension_kwargs=None,
            scenario_creator_kwargs=None,
            variable_probability=None,
            E1_tolerance=1e-5
    ):
        super().__init__(
            options,
            all_scenario_names,
            scenario_creator,
            scenario_denouement=scenario_denouement,
            all_nodenames=all_nodenames,
            mpicomm=mpicomm,
            scenario_creator_kwargs=scenario_creator_kwargs,
            variable_probability=variable_probability,
        )
        self.current_solver_options = None
        self.extensions = extensions
        self.extension_kwargs = extension_kwargs

        if (self.extensions is not None):
            if self.extension_kwargs is None:
                self.extobject = self.extensions(self)
            else:
                self.extobject = self.extensions(
                    self, **self.extension_kwargs
                )


    def solve_one(self, solver_options, k, s,
                  dtiming=False,
                  gripe=False,
                  tee=False,
                  verbose=False,
                  disable_pyomo_signal_handling=False,
                  update_objective=True):
        """ Solve one subproblem.

        Args:
            solver_options (dict or None): 
                The scenario solver options.
            k (str): 
                Subproblem name.
            s (ConcreteModel with appendages): 
                The subproblem to solve.
            dtiming (boolean, optional): 
                If True, reports timing values. Default False.
            gripe (boolean, optional):
                If True, outputs a message when a solve fails. Default False.
            tee (boolean, optional):
                If True, displays solver output. Default False.
            verbose (boolean, optional):
                If True, displays verbose output. Default False.
            disable_pyomo_signal_handling (boolean, optional):
                True for asynchronous PH; ignored for persistent solvers.
                Default False.
            update_objective (boolean, optional):
                If True, and a persistent solver is used, update
                the persistent solver's objective

        Returns:
            float:
                Pyomo solve time in seconds.
        """


        def _vb(msg): 
            if verbose and self.cylinder_rank == 0:
                print ("(rank0) " + msg)
        
        # if using a persistent solver plugin,
        # re-compile the objective due to changed weights and x-bars
        if update_objective and (sputils.is_persistent(s._solver_plugin)):
            set_objective_start_time = time.time()

            active_objective_datas = list(s.component_data_objects(
                pyo.Objective, active=True, descend_into=True))
            if len(active_objective_datas) > 1:
                raise RuntimeError('Multiple active objectives identified '
                                   'for scenario {sn}'.format(sn=s._name))
            elif len(active_objective_datas) < 1:
                raise RuntimeError('Could not find any active objectives '
                                   'for scenario {sn}'.format(sn=s._name))
            else:
                s._solver_plugin.set_objective(active_objective_datas[0])

            if dtiming:

                set_objective_time = time.time() - set_objective_start_time

                all_set_objective_times = self.mpicomm.gather(set_objective_time,
                                                          root=0)
                if self.cylinder_rank == 0:
                    print("Set objective times (seconds):")
                    print("\tmin=%4.2f mean=%4.2f max=%4.2f" %
                          (np.mean(all_set_objective_times),
                           np.mean(all_set_objective_times),
                           np.max(all_set_objective_times)))

        if self.extensions is not None:
            results = self.extobject.pre_solve(s)

        solve_start_time = time.time()
        if (solver_options):
            _vb("Using sub-problem solver options="
                + str(solver_options))
            for option_key,option_value in solver_options.items():
                s._solver_plugin.options[option_key] = option_value

        solve_keyword_args = dict()
        if self.cylinder_rank == 0:
            if tee is not None and tee is True:
                solve_keyword_args["tee"] = True
        if (sputils.is_persistent(s._solver_plugin)):
            solve_keyword_args["save_results"] = False
        elif disable_pyomo_signal_handling:
            solve_keyword_args["use_signal_handling"] = False

        try:
            results = s._solver_plugin.solve(s,
                                             **solve_keyword_args,
                                             load_solutions=False)
            solver_exception = None
        except Exception as e:
            results = None
            solver_exception = e

        if self.extensions is not None:
            results = self.extobject.post_solve(s, results)

        pyomo_solve_time = time.time() - solve_start_time
        if (results is None) or (len(results.solution) == 0) or \
                (results.solution(0).status == SolutionStatus.infeasible) or \
                (results.solver.termination_condition == TerminationCondition.infeasible) or \
                (results.solver.termination_condition == TerminationCondition.infeasibleOrUnbounded) or \
                (results.solver.termination_condition == TerminationCondition.unbounded):

            s._mpisppy_data.scenario_feasible = False

            if gripe:
                name = self.__class__.__name__
                if self.spcomm:
                    name = self.spcomm.__class__.__name__
                print (f"[{name}] Solve failed for scenario {s.name}")
                if results is not None:
                    print ("status=", results.solver.status)
                    print ("TerminationCondition=",
                           results.solver.termination_condition)

            if solver_exception is not None:
                raise solver_exception

        else:
            if sputils.is_persistent(s._solver_plugin):
                s._solver_plugin.load_vars()
            else:
                s.solutions.load_from(results)
            if self.is_minimizing:
                s._mpisppy_data.outer_bound = results.Problem[0].Lower_bound
            else:
                s._mpisppy_data.outer_bound = results.Problem[0].Upper_bound
            s._mpisppy_data.scenario_feasible = True
        # TBD: get this ready for IPopt (e.g., check feas_prob every time)
        # propogate down
        if self.bundling: # must be a bundle
            for sname in s._ef_scenario_names:
                 self.local_scenarios[sname]._mpisppy_data.scenario_feasible\
                     = s._mpisppy_data.scenario_feasible
        return pyomo_solve_time


    def solve_loop(self, solver_options=None,
                   use_scenarios_not_subproblems=False,
                   dtiming=False,
                   gripe=False,
                   disable_pyomo_signal_handling=False,
                   tee=False,
                   verbose=False):
        """ Loop over `local_subproblems` and solve them in a manner 
        dicated by the arguments. 
        
        In addition to changing the Var values in the scenarios, this function
        also updates the `_PySP_feas_indictor` to indicate which scenarios were
        feasible/infeasible.

        Args:
            solver_options (dict, optional):
                The scenario solver options.
            use_scenarios_not_subproblems (boolean, optional):
                If True, solves individual scenario problems, not subproblems.
                This distinction matters when using bundling. Default is False.
            dtiming (boolean, optional):
                If True, reports solve timing information. Default is False.
            gripe (boolean, optional):
                If True, output a message when a solve fails. Default is False.
            disable_pyomo_signal_handling (boolean, optional):
                True for asynchronous PH; ignored for persistent solvers.
                Default False.
            tee (boolean, optional):
                If True, displays solver output. Default False.
            verbose (boolean, optional):
                If True, displays verbose output. Default False.
        """

        """ Developer notes:

        This function assumes that every scenario already has a
        `_solver_plugin` attached.

        I am not sure what happens with solver_options None for a persistent
        solver. Do options persist?

        set_objective takes care of W and prox changes.
        """
        def _vb(msg): 
            if verbose and self.cylinder_rank == 0:
                print ("(rank0) " + msg)
        _vb("Entering solve_loop function.")
        logger.debug("  early solve_loop for rank={}".format(self.cylinder_rank))

        # note that when there is no bundling, scenarios are subproblems
        if use_scenarios_not_subproblems:
            s_source = self.local_scenarios
        else:
            s_source = self.local_subproblems
        for k,s in s_source.items():
            logger.debug("  in loop solve_loop k={}, rank={}".format(k, self.cylinder_rank))
            if tee:
                print(f"Tee solve for {k} on global rank {self.global_rank}")
            pyomo_solve_time = self.solve_one(solver_options, k, s,
                                              dtiming=dtiming,
                                              verbose=verbose,
                                              tee=tee,
                                              gripe=gripe,
                disable_pyomo_signal_handling=disable_pyomo_signal_handling
            )

        if dtiming:
            all_pyomo_solve_times = self.mpicomm.gather(pyomo_solve_time, root=0)
            if self.cylinder_rank == 0:
                print("Pyomo solve times (seconds):")
                print("\tmin=%4.2f mean=%4.2f max=%4.2f" %
                      (np.min(all_pyomo_solve_times),
                      np.mean(all_pyomo_solve_times),
                      np.max(all_pyomo_solve_times)))


    def Eobjective(self, verbose=False):
        """ Compute the expected objective function across all scenarios.

        Note: 
            Assumes the optimization is done beforehand,
            therefore DOES NOT CHECK FEASIBILITY or NON-ANTICIPATIVITY!
            This method uses whatever the current value of the objective
            function is.

        Args:
            verbose (boolean, optional):
                If True, displays verbose output. Default False.

        Returns:
            float:
                The expected objective function value
        """
        local_Eobjs = []
        for k,s in self.local_scenarios.items():
            if self.bundling:
                objfct = self.saved_objs[k]
            else:
                objfct = sputils.find_active_objective(s)
            local_Eobjs.append(s._mpisppy_probability * pyo.value(objfct))
            if verbose:
                print ("caller", inspect.stack()[1][3])
                print ("E_Obj Scenario {}, prob={}, Obj={}, ObjExpr={}"\
                       .format(k, s._mpisppy_probability, pyo.value(objfct), objfct.expr))

        local_Eobj = np.array([math.fsum(local_Eobjs)])
        global_Eobj = np.zeros(1)
        self.mpicomm.Allreduce(local_Eobj, global_Eobj, op=MPI.SUM)

        return global_Eobj[0]


    def Ebound(self, verbose=False, extra_sum_terms=None):
        """ Compute the expected outer bound across all scenarios.

        Note: 
            Assumes the optimization is done beforehand.
            Uses whatever bound is currently  attached to the subproblems.

        Args:
            verbose (boolean):
                If True, displays verbose output. Default False.
            extra_sum_terms: (None or iterable)
                If iterable, additional terms to put in the floating-point
                sum reduction

        Returns:
            float:
                The expected objective outer bound.
        """
        local_Ebounds = []
        for k,s in self.local_subproblems.items():
            logger.debug("  in loop Ebound k={}, rank={}".format(k, self.cylinder_rank))
            local_Ebounds.append(s._mpisppy_probability * s._mpisppy_data.outer_bound)
            if verbose:
                print ("caller", inspect.stack()[1][3])
                print ("E_Bound Scenario {}, prob={}, bound={}"\
                       .format(k, s._mpisppy_probability, s._mpisppy_data.outer_bound))

        if extra_sum_terms is not None:
            local_Ebound_list = [math.fsum(local_Ebounds)] + list(extra_sum_terms)
        else:
            local_Ebound_list = [math.fsum(local_Ebounds)]

        local_Ebound = np.array(local_Ebound_list)
        global_Ebound = np.zeros(len(local_Ebound_list))
        
        self.mpicomm.Allreduce(local_Ebound, global_Ebound, op=MPI.SUM)

        if extra_sum_terms is None:
            return global_Ebound[0]
        else:
            return global_Ebound[0], global_Ebound[1:]


    def _update_E1(self):
        """ Add up the probabilities of all scenarios using a reduce call.
            then attach it to the PH object as a float.
        """
        localP = np.zeros(1, dtype='d')
        globalP = np.zeros(1, dtype='d')

        for k,s in self.local_scenarios.items():
            localP[0] +=  s._mpisppy_probability

        self.mpicomm.Allreduce([localP, MPI.DOUBLE],
                           [globalP, MPI.DOUBLE],
                           op=MPI.SUM)

        self.E1 = float(globalP[0])


    def feas_prob(self):
        """ Compute the total probability of all feasible scenarios.

        This function can be used to check whether all scenarios are feasible
        by comparing the return value to one.
        
        Note:
            This function assumes the scenarios have a boolean
            `_mpisppy_data.scenario_feasible` attribute.

        Returns:
            float:
                Sum of the scenario probabilities over all feasible scenarios.
                This value equals E1 if all scenarios are feasible.
        """

        # locals[0] is E_feas and locals[1] is E_1
        locals = np.zeros(1, dtype='d')
        globals = np.zeros(1, dtype='d')

        for k,s in self.local_scenarios.items():
            if s._mpisppy_data.scenario_feasible:
                locals[0] += s._mpisppy_probability

        self.mpicomm.Allreduce([locals, MPI.DOUBLE],
                           [globals, MPI.DOUBLE],
                           op=MPI.SUM)

        return float(globals[0])


    def infeas_prob(self):
        """ Sum the total probability for all infeasible scenarios.

        Note:
            This function assumes the scenarios have a boolean
            `_mpisppy_data.scenario_feasible` attribute.

        Returns:
            float:
                Sum of the scenario probabilities over all infeasible scenarios.
                This value equals 0 if all scenarios are feasible.
        """

        locals = np.zeros(1, dtype='d')
        globals = np.zeros(1, dtype='d')

        for k,s in self.local_scenarios.items():
            if not s._mpisppy_data.scenario_feasible:
                locals[0] += s._mpisppy_probability

        self.mpicomm.Allreduce([locals, MPI.DOUBLE],
                           [globals, MPI.DOUBLE],
                           op=MPI.SUM)

        return float(globals[0])


    def avg_min_max(self, compstr):
        """ Can be used to track convergence progress.

        Args:
            compstr (str): 
                The name of the Pyomo component. Should not be indexed.

        Returns:
            tuple: 
                Tuple containing

                avg (float): 
                    Average across all scenarios.
                min (float):
                    Minimum across all scenarios.
                max (float):
                    Maximum across all scenarios.

        Note:
            WARNING: Does a Allreduce.
            Not user-friendly. If you give a bad compstr, it will just crash.
        """
        firsttime = True
        localavg = np.zeros(1, dtype='d')
        localmin = np.zeros(1, dtype='d')
        localmax = np.zeros(1, dtype='d')
        globalavg = np.zeros(1, dtype='d')
        globalmin = np.zeros(1, dtype='d')
        globalmax = np.zeros(1, dtype='d')

        v_cuid = pyo.ComponentUID(compstr)

        for k,s in self.local_scenarios.items():

            compv = pyo.value(v_cuid.find_component_on(s))

            
            ###compv = pyo.value(getattr(s, compstr))
            localavg[0] += s._mpisppy_probability * compv  
            if compv < localmin[0] or firsttime:
                localmin[0] = compv
            if compv > localmax[0] or firsttime:
                localmax[0] = compv
            firsttime = False

        self.comms["ROOT"].Allreduce([localavg, MPI.DOUBLE],
                                     [globalavg, MPI.DOUBLE],
                                     op=MPI.SUM)
        self.comms["ROOT"].Allreduce([localmin, MPI.DOUBLE],
                                     [globalmin, MPI.DOUBLE],
                                     op=MPI.MIN)
        self.comms["ROOT"].Allreduce([localmax, MPI.DOUBLE],
                                     [globalmax, MPI.DOUBLE],
                                     op=MPI.MAX)
        return (float(globalavg[0]),
                float(globalmin[0]),
                float(globalmax[0]))


    def _put_nonant_cache(self, cache):
        """ Put the value in the cache for noants *for all local scenarios*
        Args:
            cache (np vector) to receive the nonant's for all local scenarios

        """
        ci = 0 # Cache index
        for sname, model in self.local_scenarios.items():
            if model._mpisppy_data.nonant_cache is None:
                raise RuntimeError(f"Rank {self.global_rank} Scenario {sname}"
                                   " nonant_cache is None"
                                   " (call _save_nonants first?)")
            for i,_ in enumerate(model._mpisppy_data.nonant_indices):
                assert(ci < len(cache))
                model._mpisppy_data.nonant_cache[i] = cache[ci]
                ci += 1


    def _restore_original_fixedness(self):
        # We are going to hack a little to get the original fixedness, but current values
        # (We are assuming that algorithms are not fixing anticipative vars; but if they
        # do, they had better put their fixedness back to its correct state.)
        self._save_nonants()
        for k,s in self.local_scenarios.items():        
            for ci, _ in enumerate(s._mpisppy_data.nonant_indices):
                s._mpisppy_data.fixedness_cache[ci] = s._mpisppy_data.original_fixedness[ci]
        self._restore_nonants()


    def _fix_nonants(self, cache):
        """ Fix the Vars subject to non-anticipativity at given values.
            Loop over the scenarios to restore, but loop over subproblems
            to alert persistent solvers.
        Args:
            cache (ndn dict of list or numpy vector): values at which to fix
        WARNING: 
            We are counting on Pyomo indices not to change order between
            when the cache_list is created and used.
        NOTE:
            You probably want to call _save_nonants right before calling this
        """
        for k,s in self.local_scenarios.items():

            persistent_solver = None
            if (sputils.is_persistent(s._solver_plugin)):
                persistent_solver = s._solver_plugin

            nlens = s._mpisppy_data.nlens
            for node in s._mpisppy_node_list:
                ndn = node.name
                if ndn not in cache:
                    raise RuntimeError("Could not find {} in {}"\
                                       .format(ndn, cache))
                if cache[ndn] is None:
                    raise RuntimeError("Empty cache for scen={}, node={}".format(k, ndn))
                if len(cache[ndn]) != nlens[ndn]:
                    raise RuntimeError("Needed {} nonant Vars for {}, got {}"\
                                       .format(nlens[ndn], ndn, len(cache[ndn])))
                for i in range(nlens[ndn]): 
                    this_vardata = node.nonant_vardata_list[i]
                    this_vardata._value = cache[ndn][i]
                    this_vardata.fix()
                    if persistent_solver is not None:
                        persistent_solver.update_var(this_vardata)


    def _restore_nonants(self):
        """ Restore nonanticipative variables to their original values.
            
        This function works in conjunction with _save_nonants. 
        
        We loop over the scenarios to restore variables, but loop over
        subproblems to alert persistent solvers.

        Warning: 
            We are counting on Pyomo indices not to change order between save
            and restoration. THIS WILL NOT WORK ON BUNDLES (Feb 2019) but
            hopefully does not need to.
        """
        for k,s in self.local_scenarios.items():

            persistent_solver = None
            if (sputils.is_persistent(s._solver_plugin)):
                persistent_solver = s._solver_plugin

            for ci, vardata in enumerate(s._mpisppy_data.nonant_indices.values()):
                vardata._value = s._mpisppy_data.nonant_cache[ci]
                vardata.fixed = s._mpisppy_data.fixedness_cache[ci]

                if persistent_solver is not None:
                    persistent_solver.update_var(vardata)


    def _save_nonants(self):
        """ Save the values and fixedness status of the Vars that are
        subject to non-anticipativity.

        Note:
            Assumes nonant_cache is on the scenarios and can be used
            as a list, or puts it there.
        Warning: 
            We are counting on Pyomo indices not to change order before the
            restoration. We also need the Var type to remain stable.
        Note:
            The value cache is np because it might be transmitted
        """
        for k,s in self.local_scenarios.items():
            nlens = s._mpisppy_data.nlens
            if not hasattr(s._mpisppy_data,"nonant_cache"):
                clen = sum(nlens[ndn] for ndn in nlens)
                s._mpisppy_data.nonant_cache = np.zeros(clen, dtype='d')
                s._mpisppy_data.fixedness_cache = [None for _ in range(clen)]

            for ci, xvar in enumerate(s._mpisppy_data.nonant_indices.values()):
                s._mpisppy_data.nonant_cache[ci]  = xvar._value
                s._mpisppy_data.fixedness_cache[ci]  = xvar.is_fixed()


    def _save_original_nonants(self):
        """ Save the current value of the nonanticipative variables.
            
        Values are saved in the `_PySP_original_nonants` attribute. Whether
        the variable was fixed is stored in `_PySP_original_fixedness`.
        """
        for k,s in self.local_scenarios.items():
            if hasattr(s,"_PySP_original_fixedness"):
                print ("ERROR: Attempt to replace original nonants")
                raise
            if not hasattr(s._mpisppy_data,"nonant_cache"):
                # uses nonant cache to signal other things have not
                # been created 
                # TODO: combine cache creation (or something else)
                clen = len(s._mpisppy_data.nonant_indices)
                s._mpisppy_data.original_fixedness = [None] * clen
                s._mpisppy_data.original_nonants = np.zeros(clen, dtype='d')

            for ci, xvar in enumerate(s._mpisppy_data.nonant_indices.values()):
                s._mpisppy_data.original_fixedness[ci]  = xvar.is_fixed()
                s._mpisppy_data.original_nonants[ci]  = xvar._value


    def _restore_original_nonants(self):
        """ Restore nonanticipative variables to their original values.
            
        This function works in conjunction with _save_original_nonants. 
        
        We loop over the scenarios to restore variables, but loop over
        subproblems to alert persistent solvers.

        Warning: 
            We are counting on Pyomo indices not to change order between save
            and restoration. THIS WILL NOT WORK ON BUNDLES (Feb 2019) but
            hopefully does not need to.
        """
        for k,s in self.local_scenarios.items():

            persistent_solver = None
            if not self.bundling:
                if (sputils.is_persistent(s._solver_plugin)):
                    persistent_solver = s._solver_plugin
            else:
                print("restore_original_nonants called for a bundle")
                raise

            for ci, vardata in enumerate(s._mpisppy_data.nonant_indices.values()):
                vardata._value = s._mpisppy_data.original_nonants[ci]
                vardata.fixed = s._mpisppy_data.original_fixedness[ci]
                if persistent_solver != None:
                    persistent_solver.update_var(vardata)


    def FormEF(self, scen_dict, EF_name=None):
        """ Make the EF for a list of scenarios. 
        
        This function is mainly to build bundles. To build (and solve) the
        EF of the entire problem, use the EF class instead.

        Args:
            scen_dict (dict): 
                Subset of local_scenarios; the scenarios to put in the EF. THe
                dictionary maps sccneario names (strings) to scenarios (Pyomo
                concrete model objects).
            EF_name (string, optional):
                Name for the resulting EF model.

        Returns:
            :class:`pyomo.environ.ConcreteModel`: 
                The EF with explicit non-anticipativity constraints.

        Raises:
            RuntimeError:
                If the `scen_dict` is empty, or one of the scenarios in
                `scen_dict` is not owned locally (i.e. is not in
                `local_scenarios`).

        Note: 
            We attach a list of the scenario names called _PySP_subsecen_names
        Note:
            We deactivate the objective on the scenarios.
        Note:
            The scenarios are sub-blocks, so they naturally get the EF solution
            Also the EF objective references Vars and Parms on the scenarios
            and hence is automatically updated when the scenario
            objectives are. THIS IS ALL CRITICAL to bundles.
            xxxx TBD: ask JP about objective function transmittal to persistent solvers
        Note:
            Objectives are scaled (normalized) by _mpisppy_probability
        """
        if len(scen_dict) == 0:
            raise RuntimeError("Empty scenario list for EF")

        if len(scen_dict) == 1:
            sname, scenario_instance = list(scen_dict.items())[0]
            if EF_name is not None:
                print ("WARNING: EF_name="+EF_name+" not used; singleton="+sname)
                print ("MAJOR WARNING: a bundle of size one encountered; if you try to compute bounds it might crash (Feb 2019)")
            return scenario_instance

        # The individual scenario instances are sub-blocks of the binding
        # instance. Needed to facilitate bundles + persistent solvers
        if not hasattr(self, "saved_objs"): # First bundle
             self.saved_objs = dict()

        for sname, scenario_instance in scen_dict.items():
            if sname not in self.local_scenarios:
                raise RuntimeError("EF scen not in local_scenarios="+sname)
            self.saved_objs[sname] = sputils.find_active_objective(scenario_instance)

        EF_instance = sputils._create_EF_from_scen_dict(scen_dict, EF_name=EF_name,
                        nonant_for_fixed_vars=False)
        return EF_instance


    def subproblem_creation(self, verbose=False):
        """ Create local subproblems (not local scenarios).

        If bundles are specified, this function creates the bundles.
        Otherwise, this function simply copies pointers to the already-created
        `local_scenarios`.

        Args:
            verbose (boolean, optional):
                If True, displays verbose output. Default False.
        """
        self.local_subproblems = dict()
        if self.bundling:
            rank_local = self.cylinder_rank
            for bun in self.names_in_bundles[rank_local]:
                sdict = dict()
                bname = "rank" + str(self.cylinder_rank) + "bundle" + str(bun)
                for sname in self.names_in_bundles[rank_local][bun]:
                    if (verbose and self.cylinder_rank==0):
                        print ("bundling "+sname+" into "+bname)
                    scen = self.local_scenarios[sname]
                    scen._mpisppy_data.bundlename = bname
                    sdict[sname] = scen
                self.local_subproblems[bname] = self.FormEF(sdict, bname)
                self.local_subproblems[bname].scen_list = \
                    self.names_in_bundles[rank_local][bun]
                self.local_subproblems[bname]._mpisppy_probability = \
                                    sum(s._mpisppy_probability for s in sdict.values())
        else:
            for sname, s in self.local_scenarios.items():
                self.local_subproblems[sname] = s
                self.local_subproblems[sname].scen_list = [sname]


    def _create_solvers(self):

        for sname, s in self.local_subproblems.items(): # solver creation
            s._solver_plugin = SolverFactory(self.options["solvername"])

            if (sputils.is_persistent(s._solver_plugin)):
                dtiming = ("display_timing" in self.options) and self.options["display_timing"]
                if dtiming:
                    set_instance_start_time = time.time()

                # this loop is required to address the sitution where license
                # token servers become temporarily over-subscribed / non-responsive
                # when large numbers of ranks are in use.

                # these parameters should eventually be promoted to a non-PH
                # general class / location. even better, the entire retry
                # logic can be encapsulated in a sputils.py function.
                MAX_ACQUIRE_LICENSE_RETRY_ATTEMPTS = 5
                LICENSE_RETRY_SLEEP_TIME = 2 # in seconds
        
                num_retry_attempts = 0
                while True:
                    try:
                        s._solver_plugin.set_instance(s)
                        if num_retry_attempts > 0:
                            print("Acquired solver license (call to set_instance() for scenario=%s) after %d retry attempts" % (sname, num_retry_attempts))
                        break
                    # pyomo presently has no general way to trap a license acquisition
                    # error - so we're stuck with trapping on "any" exception. not ideal.
                    except:
                        if num_retry_attempts == 0:
                            print("Failed to acquire solver license (call to set_instance() for scenario=%s) after first attempt" % (sname))
                        else:
                            print("Failed to acquire solver license (call to set_instance() for scenario=%s) after %d retry attempts" % (sname, num_retry_attempts))
                        if num_retry_attempts == MAX_ACQUIRE_LICENSE_RETRY_ATTEMPTS:
                            raise RuntimeError("Failed to acquire solver license - call to set_instance() for scenario=%s failed after %d retry attempts" % (sname, num_retry_attempts))
                        else:
                            print("Sleeping for %d seconds before re-attempting" % LICENSE_RETRY_SLEEP_TIME)
                            time.sleep(LICENSE_RETRY_SLEEP_TIME)
                            num_retry_attempts += 1

                if dtiming:
                    set_instance_time = time.time() - set_instance_start_time
                    all_set_instance_times = self.mpicomm.gather(set_instance_time,
                                                                 root=0)
                    if self.cylinder_rank == 0:
                        print("Set instance times:")
                        print("\tmin=%4.2f mean=%4.2f max=%4.2f" %
                              (np.min(all_set_instance_times),
                               np.mean(all_set_instance_times),
                               np.max(all_set_instance_times)))

            ## if we have bundling, attach
            ## the solver plugin to the scenarios
            ## as well to avoid some gymnastics
            if self.bundling:
                for scen_name in s.scen_list:
                    scen = self.local_scenarios[scen_name]
                    scen._solver_plugin = s._solver_plugin



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' Base class for converger objects

    DTM Dec 2019

    Replaces the old implementation in which
    convergers were modules rather than classes.
'''

import abc

class Converger:
    ''' Abstract base class for converger monitors.

        Args:
            opt (SPBase): The SPBase object for the current model
    '''
    def __init__(self, opt):
        pass

    @abc.abstractmethod
    def is_converged(self):
        ''' Indicated whether the algorithm has converged.
            
            Must return a boolean. If True, the algorithm will terminate at the
            current iteration--no more solves will be performed by SPBase.
            Otherwise, the iterations will continue.
        '''
        pass



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# Indepedent Lagrangian that takes x values as input and
# updates its own W.

import time
import mpisppy.cylinders.spoke

class LagrangerOuterBound(mpisppy.cylinders.spoke.OuterBoundNonantSpoke):

    converger_spoke_char = 'A'

    def lagrangian_prep(self):
        verbose = self.opt.options['verbose']
        # Scenarios are created here
        self.opt.PH_Prep(attach_prox=False)
        self.opt._reenable_W()
        self.opt.subproblem_creation(verbose)
        self.opt._create_solvers()
        if "lagranger_rho_rescale_factors_json" in self.opt.options and\
            self.opt.options["lagranger_rho_rescale_factors_json"] is not None:
            with open(args.lagranger_rho_rescale_factors_json, "r") as fin:
                din = json.load(fin)
            self.rho_rescale_factors = {int(i): din[i] for i in din}
        else:
            self.rho_rescale_factors = None
        # side-effect is needed: create the nonant_cache
        self.opt._save_nonants()

    def _lagrangian(self, iternum):
        verbose = self.opt.options['verbose']
        # see if rho should be rescaled
        if self.rho_rescale_factors is not None\
           and iternum in self.rho_rescale_factors:
            _rescale_rho(self.rho_rescale_factors[iternum])
        teeme = False
        if "tee-rank0-solves" in self.opt.options and self.opt.cylinder_rank == 0:
            teeme = self.opt.options['tee-rank0-solves']

        self.opt.solve_loop(
            solver_options=self.opt.current_solver_options,
            dtiming=False,
            gripe=True,
            tee=teeme,
            verbose=verbose
        )

        # Compute the resulting bound
        return self.opt.Ebound(verbose)


    def _rescale_rho(rf):
        # IMPORTANT: the scalings accumulate.
        # E.g., 0.5 then 2.0 gets you back where you started.
        for (sname, scenario) in self.local_scenarios.items():
            for ndn_i, xvar in scenario._mpisppy_data.nonant_indices.items():
                scenario._mpisppy_model.rho[ndn_i] *= rf
        
    
    def _update_weights_and_solve(self, iternum):
        # Work with the nonants that we have (and we might not have any yet).
        self.opt._put_nonant_cache(self.localnonants)
        self.opt._restore_nonants()
        verbose = self.opt.options["verbose"]
        self.opt.Compute_Xbar(verbose=verbose)
        self.opt.Update_W(verbose=verbose)
        return self._lagrangian(iternum)

    def main(self):
        # The rho_setter should be attached to the opt object
        rho_setter = None
        if hasattr(self.opt, 'rho_setter'):
            rho_setter = self.opt.rho_setter

        self.lagrangian_prep()

        self.trivial_bound = self._lagrangian(0)

        self.bound = self.trivial_bound

        self.A_iter = 1
        while not self.got_kill_signal():
            # because of aph, do not check for new data, just go for it
            self.bound = self._update_weights_and_solve(self.A_iter)
            self.A_iter += 1

    def finalize(self):
        '''
        Do one final lagrangian pass with the final
        PH weights. Useful for when PH convergence
        and/or iteration limit is the cause of termination
        '''
        self.final_bound = self._update_weights_and_solve(self.A_iter)
        self.bound = self.final_bound
        return self.final_bound



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# udpated April 20
# specific xhat supplied (copied from xhatlooper_bounder by DLW, Dec 2019)

import mpisppy.cylinders.spoke as spoke
from mpisppy.extensions.xhatspecific import XhatSpecific
from mpisppy.utils.xhat_eval import Xhat_Eval

import mpi4py.MPI as mpi
import logging

fullcomm = mpi.COMM_WORLD
global_rank = fullcomm.Get_rank()
fullcom_n_proc = fullcomm.Get_size()


############################################################################
class XhatSpecificInnerBound(spoke.InnerBoundNonantSpoke):

    converger_spoke_char = 'S'

    def ib_prep(self):
        """
        Set up the objects needed for bounding.

        Returns:
            xhatter (xhatspecific object): Constructed by a call to Prep
        """
        if "bundles_per_rank" in self.opt.options\
           and self.opt.options["bundles_per_rank"] != 0:
            raise RuntimeError("xhat spokes cannot have bundles (yet)")

        if not isinstance(self.opt, Xhat_Eval):
            raise RuntimeError("XhatShuffleInnerBound must be used with Xhat_Eval.")

        verbose = self.opt.options['verbose']
        xhatter = XhatSpecific(self.opt)
        # somehow deal with the prox option .... TBD .... important for aph APH

        # begin iter0 stuff
        xhatter.pre_iter0()
        self.opt._save_original_nonants()

        teeme = False
        if ("tee-rank0-solves" in self.opt.options):
            teeme = self.opt.options['tee-rank0-solves']

        self.opt.solve_loop(solver_options=self.opt.current_solver_options,
                            dtiming=False,
                            gripe=True,
                            tee=teeme,
                            verbose=verbose)
        self.opt._update_E1()  # Apologies for doing this after the iter0 solves...
        if (abs(1 - self.opt.E1) > self.opt.E1_tolerance):
            if self.opt.cylinder_rank == 0:
                print("ERROR")
                print("Total probability of scenarios was ", self.opt.E1)
                print("E1_tolerance = ", self.opt.E1_tolerance)
            quit()
        infeasP = self.opt.infeas_prob()
        if infeasP != 0.:
            if self.opt.cylinder_rank == 0:
                print("ERROR")
                print("Infeasibility detected; E_infeas, E1=", infeasP, self.opt.E1)
            quit()

        ### end iter0 stuff

        xhatter.post_iter0()
        self.opt._save_nonants()  # make the cache

        return xhatter

    def main(self):
        """
        Entry point. Communicates with the optimization companion.

        """
        dtm = logging.getLogger(f'dtm{global_rank}')
        verbose = self.opt.options["verbose"] # typing aid  
        logging.debug("Enter xhatspecific main on rank {}".format(global_rank))

        # What to try does not change, but the data in the scenarios should
        xhat_scenario_dict = self.opt.options["xhat_specific_options"]\
                                             ["xhat_scenario_dict"]

        xhatter = self.ib_prep()

        ib_iter = 1  # ib is for inner bound
        got_kill_signal = False
        while (not self.got_kill_signal()):
            logging.debug('   IB loop iter={} on global rank {}'.\
                          format(ib_iter, global_rank))
            # _log_values(ib_iter, self._locals, dtm)

            logging.debug('   IB got from opt on global rank {}'.\
                          format(global_rank))
            if (self.new_nonants):
                logging.debug('  and its new! on global rank {}'.\
                              format(global_rank))
                logging.debug('  localnonants={}'.format(str(self.localnonants)))

                self.opt._put_nonant_cache(self.localnonants)  # don't really need all caches
                self.opt._restore_nonants()
                innerbound = xhatter.xhat_tryit(xhat_scenario_dict, restore_nonants=False)

                self.update_if_improving(innerbound)

            ib_iter += 1

        dtm.debug(f'IB specific thread ran {ib_iter} iterations\n')
    
if __name__ == "__main__":
    print("no main.")



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
""" Code for a mipgap schedule. This can be used
    as the only extension, but it could also be called from a "multi"
    extension.
"""

import pyomo.environ as pyo
import mpisppy.extensions.extension

class Gapper(mpisppy.extensions.extension.Extension):

    def __init__(self, ph):
        self.ph = ph
        self.cylinder_rank = self.ph.cylinder_rank
        self.gapperoptions = self.ph.options["gapperoptions"] # required
        self.mipgapdict = self.gapperoptions["mipgapdict"]
        self.verbose = self.ph.options["verbose"] \
                       or self.gapperoptions["verbose"]
                       
    def _vb(self, str):
        if self.verbose and self.cylinder_rank == 0:
            print ("(rank0) mipgapper:" + str)

    def set_mipgap(self, mipgap):
        """ set the mipgap
        Args:
            float (mipgap): the gap to set
        """
        oldgap = None
        if "mipgap" in self.ph.current_solver_options:
            oldgap = self.ph.current_solver_options["mipgap"]
        self._vb("Changing mipgap from "+str(oldgap)+" to "+str(mipgap))
        self.ph.current_solver_options["mipgap"] = float(mipgap)
        
    def pre_iter0(self):
        if self.mipgapdict is None:
            return
        if 0 in self.mipgapdict:
            self.set_mipgap(self.mipgapdict[0])
                                        
    def post_iter0(self):
        return

    def miditer(self):
        if self.mipgapdict is None:
            return
        PHIter = self.ph._PHIter
        if PHIter in self.mipgapdict:
            self.set_mipgap(self.mipgapdict[PHIter])


    def enditer(self):
        return

    def post_everything(self):
        return



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# An extension to compute and output avg, min, max for
# a component (e.g., first stage cost).
# DLW, Feb 2019
# This extension uses options["avgminmax_name"]

import mpisppy.extensions.xhatbase

class MinMaxAvg(mpisppy.extensions.xhatbase.XhatBase):
    """
    Args:
        ph (PH object): the calling object
        rank (int): mpi process rank of currently running process
    """
    def __init__(self, ph, rank, n_proc):
        super().__init__(ph, rank, n_proc)
        self.compstr = self.ph.options["avgminmax_name"]

    def pre_iter0(self):
        return

    def post_iter0(self):
        avgv, minv, maxv = self.ph.avg_min_max(self.compstr)
        if (self.cylinder_rank == 0):
            print ("  ### ", self.compstr,": avg, min, max, max-min", avgv, minv, maxv, maxv-minv)
        
    def miditer(self, PHIter, conv):
        return

    def enditer(self, PHIter):
        avgv, minv, maxv = self.ph.avg_min_max(self.compstr)
        if (self.cylinder_rank == 0):
            print ("  ### ", self.compstr,": avg, min, max, max-min", avgv, minv, maxv, maxv-minv)

    def post_everything(self, PHIter, conv):
        return





# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import logging
import time
import random
import mpisppy.log
import mpisppy.utils.sputils as sputils
import mpisppy.cylinders.spoke as spoke

from math import inf
from pyomo.opt import SolverFactory, SolverStatus, TerminationCondition
from mpisppy.phbase import PHBase
from mpisppy.utils.xhat_eval import Xhat_Eval

class XhatLShapedInnerBound(spoke.InnerBoundNonantSpoke):

    converger_spoke_char = 'X'

    def xhatlshaped_prep(self):
        verbose = self.opt.options['verbose']

        if not isinstance(self.opt, Xhat_Eval):
            raise RuntimeError("XhatLShapedInnerBound must be used with Xhat_Eval.")

        teeme = False
        if "tee-rank0-solves" in self.opt.options:
            teeme = self.opt.options['tee-rank0-solves']

        self.opt.solve_loop(
            solver_options=self.opt.current_solver_options,
            dtiming=False,
            gripe=True,
            tee=teeme,
            verbose=verbose
        )
        self.opt._update_E1()  # Apologies for doing this after the solves...
        if abs(1 - self.opt.E1) > self.opt.E1_tolerance:
            if self.opt.cylinder_rank == 0:
                print("ERROR")
                print("Total probability of scenarios was ", self.opt.E1)
                print("E1_tolerance = ", self.opt.E1_tolerance)
            quit()
        infeasP = self.opt.infeas_prob()
        if infeasP != 0.:
            if self.opt.cylinder_rank == 0:
                print("ERROR")
                print("Infeasibility detected; E_infeas, E1=", infeasP, self.opt.E1)
            quit()

        self.opt._save_nonants() # make the cache

        self.opt.current_solver_options = self.opt.options["iterk_solver_options"]
        ### end iter0 stuff

    def main(self):

        self.xhatlshaped_prep()
        is_minimizing = self.opt.is_minimizing

        self.ib = inf if is_minimizing else -inf

        #xh_iter = 1
        while not self.got_kill_signal():

            if self.new_nonants:
                
                self.opt._put_nonant_cache(self.localnonants)
                self.opt._restore_nonants()
                obj = self.opt.calculate_incumbent(fix_nonants=True)

                if obj is None:
                    continue

                update = (obj < self.ib) if is_minimizing else (self.ib < obj)
                if update:
                    self.bound = obj
                    self.ib = obj

        ## TODO: Save somewhere?



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' An extension to initialize PH weights and/or PH xbar values from csv files.

    To use, specify either or both of the following keys to the options dict:

        "init_W_fname" : <str filename>
        "init_Xbar_fname": <str filename>

    If neither option is specified, this extension does nothing (i.e. is
    silent--does not raise a warning/error). If only one is specified, then
    only those values are initialized. The specified files should be
    formatted as follows:

        (W values) csv with rows: scenario_name, variable_name, weight_value
        (x values) csv with rows: variable_name, variable_value

    Rows that begin with "#" are treated as comments. If the files are missing
    any values, raises an error. Extra values are ignored.

    TODO:
        Check with bundles.

    Written: DLW, July 2019
    Modified: DTM, Aug 2019
'''

import mpisppy.utils.wxbarutils
import os # For checking if files exist
import mpisppy.extensions.extension

class WXBarReader(mpisppy.extensions.extension.Extension):
    """ Extension class for reading W values
    """
    def __init__(self, ph, rank, n_proc):

        ''' Do a bunch of checking if files exist '''
        w_fname, x_fname, sep_files = None, None, False
        if ('init_separate_W_files' in ph.options):
            sep_files = ph.options['init_separate_W_files']

        if ('init_W_fname' in ph.options):
            w_fname = ph.options['init_W_fname']
            if (not os.path.exists(w_fname)):
                if (rank == 0):
                    if (sep_files):
                        print('Cannot find path', w_fname)
                    else:
                        print('Cannot find file', w_fname)
                quit()

        if ('init_Xbar_fname' in ph.options):
            x_fname = ph.options['init_Xbar_fname']
            if (not os.path.exists(x_fname)):
                if (rank == 0):
                    print('Cannot find file', x_fname)
                quit()

        if (x_fname is None and w_fname is None and rank==0):
            print('Warning: no input files provided to WXBarReader. '
                  'W and Xbar will be initialized to their default values.')

        self.PHB = ph
        self.cylinder_rank = rank
        self.w_fname = w_fname
        self.x_fname = x_fname
        self.sep_files = sep_files

    def pre_iter0(self):
        if (self.w_fname):
            mpisppy.utils.wxbarutils.set_W_from_file(
                    self.w_fname, self.PHB, self.cylinder_rank,
                    sep_files=self.sep_files)
            self.PHB._reenable_W() # This makes a big difference.
        if (self.x_fname):
            mpisppy.utils.wxbarutils.set_xbar_from_file(self.x_fname, self.PHB)
            self.PHB._reenable_prox()

    def post_iter0(self):
        pass
        
    def miditer(self, PHIter, conv):
        ''' Called before the solveloop is called '''
        pass

    def enditer(self, PHIter):
        ''' Called after the solve loop '''
        pass

    def post_everything(self, PHIter, conv):
        pass



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# Code to evaluate a given x-hat, but given as a nonant-cache
# To test: python xhat_eval.py --num-scens=3 --EF-solver-name=cplex

import inspect
import pyomo.environ as pyo
import logging
import numpy as np
import math

import mpisppy.log
import mpisppy.phbase
from mpi4py import MPI
import mpisppy.utils.sputils as sputils
import mpisppy.spopt

    
fullcomm = MPI.COMM_WORLD
global_rank = fullcomm.Get_rank()

# Could also pass, e.g., sys.stdout instead of a filename
mpisppy.log.setup_logger("mpisppy.utils.xhat_eval",
                         "xhateval.log",
                         level=logging.CRITICAL)                         
logger = logging.getLogger("mpisppy.utils.xhat_eval")

############################################################################
class Xhat_Eval(mpisppy.spopt.SPOpt):
    """ See SPOpt for list of args. """
    
    def __init__(
        self,
        options,
        all_scenario_names,
        scenario_creator,
        scenario_denouement=None,
        all_nodenames=None,
        mpicomm=None,
        scenario_creator_kwargs=None,
        variable_probability=None,
        ):
        
        super().__init__(
            options,
            all_scenario_names,
            scenario_creator,
            scenario_denouement=scenario_denouement,
            all_nodenames=all_nodenames,
            mpicomm=mpicomm,
            scenario_creator_kwargs=scenario_creator_kwargs,
            variable_probability=variable_probability,
        )
        
        self.verbose = self.options['verbose']
        
        #TODO: CHANGE THIS AFTER UPDATE
        self.PH_extensions = None
        
        self.subproblem_creation(self.verbose)
        self._create_solvers()


    


    #======================================================================
    def solve_one(self, solver_options, k, s,
                  dtiming=False,
                  gripe=False,
                  tee=False,
                  verbose=False,
                  disable_pyomo_signal_handling=False,
                  update_objective=True,
                  compute_val_at_nonant=False):
        pyomo_solve_time = super().solve_one(solver_options, k, s,
                                             dtiming=False,
                                             gripe=False,
                                             tee=False,
                                             verbose=False,
                                             disable_pyomo_signal_handling=False,
                                             update_objective=True)
        if compute_val_at_nonant:
                if self.bundling:
                    objfct = self.saved_objs[k]
                    
                else:
                    objfct = sputils.find_active_objective(s)
                    if self.verbose:
                        print ("caller", inspect.stack()[1][3])
                        print ("E_Obj Scenario {}, prob={}, Obj={}, ObjExpr={}"\
                               .format(k, s._mpisppy_probability, pyo.value(objfct), objfct.expr))
                self.objs_dict[k] = pyo.value(objfct)
        return(pyomo_solve_time)


    def solve_loop(self, solver_options=None,
                   use_scenarios_not_subproblems=False,
                   dtiming=False,
                   gripe=False,
                   disable_pyomo_signal_handling=False,
                   tee=False,
                   verbose=False,
                   compute_val_at_nonant=False):
        """ Loop over self.local_subproblems and solve them in a manner 
            dicated by the arguments. In addition to changing the Var
            values in the scenarios, update _PySP_feas_indictor for each.

        ASSUMES:
            Every scenario already has a _solver_plugin attached.

        Args:
            solver_options (dict or None): the scenario solver options
            use_scenarios_not_subproblems (boolean): for use by bounds
            dtiming (boolean): indicates that timing should be reported
            gripe (boolean): output a message if a solve fails
            disable_pyomo_signal_handling (boolean): set to true for asynch, 
                                                     ignored for persistent solvers.
            tee (boolean): show solver output to screen if possible
            verbose (boolean): indicates verbose output
            compute_val_at_nonant (boolean): indicate that self.objs_dict should
                                            be created and computed.
                                            

        NOTE: I am not sure what happens with solver_options None for
              a persistent solver. Do options persist?

        NOTE: set_objective takes care of W and prox changes.
        """
        def _vb(msg): 
            if verbose and self.cylinder_rank == 0:
                print ("(rank0) " + msg)
        logger.debug("  early solve_loop for rank={}".format(self.cylinder_rank))

        # note that when there is no bundling, scenarios are subproblems
        if use_scenarios_not_subproblems:
            s_source = self.local_scenarios
        else:
            s_source = self.local_subproblems
            
        if compute_val_at_nonant:
            self.objs_dict={}
        
        for k,s in s_source.items():
            if tee:
                print(f"Tee solve for {k} on global rank {self.global_rank}")
            logger.debug("  in loop solve_loop k={}, rank={}".format(k, self.cylinder_rank))

            pyomo_solve_time = self.solve_one(solver_options, k, s,
                                              dtiming=dtiming,
                                              verbose=verbose,
                                              tee=tee,
                                              gripe=gripe,
                disable_pyomo_signal_handling=disable_pyomo_signal_handling,
                compute_val_at_nonant=compute_val_at_nonant)

        if dtiming:
            all_pyomo_solve_times = self.mpicomm.gather(pyomo_solve_time, root=0)
            if self.cylinder_rank == 0:
                print("Pyomo solve times (seconds):")
                print("\tmin=%4.2f mean=%4.2f max=%4.2f" %
                      (np.min(all_pyomo_solve_times),
                      np.mean(all_pyomo_solve_times),
                      np.max(all_pyomo_solve_times)))
                
    #======================================================================   
    def Eobjective(self, verbose=False, fct=None):
        """ Compute the expected value of the composition of the objective function 
            and a given function fct across all scenarios.

        Note: 
            Assumes the optimization is done beforehand,
            therefore DOES NOT CHECK FEASIBILITY or NON-ANTICIPATIVITY!
            This method uses whatever the current value of the objective
            function is.

        Args:
            verbose (boolean, optional):
                If True, displays verbose output. Default False.
            
            fct (function, optional):
                A function R-->R^p, such as x|-->(x,x^2,x^3). Default is None
                If fct is None, Eobjective returns the exepected value.
                

        Returns:
            float or numpy.array:
                The expected objective function value. 
                If fct is R-->R, returns a float.
                If fct is R-->R^p with p>1, returns a np.array of length p
        """
        if fct is None:
            return super().Eobjective(verbose=verbose)
        
        if not hasattr(self, "objs_dict"):
            raise RuntimeError("Values of the objective functions for each scenario"+
                               " at xhat have to be computed before running Eobjective")
        

        local_Eobjs = []
        for k,s in self.local_scenarios.items():
            if not k in self.objs_dict:
                raise RuntimeError(f"No value has been calculated for the scenario {k}")
            local_Eobjs.append(s._mpisppy_probability * fct(self.objs_dict[k]))
        local_Eobjs = np.array(local_Eobjs)
        local_Eobj = np.array([np.sum(local_Eobjs,axis=0)])
        global_Eobj = np.zeros(len(local_Eobj))
        self.mpicomm.Allreduce(local_Eobj, global_Eobj, op=MPI.SUM)
        if len(global_Eobj)==1:
            global_Eobj = global_Eobj[0]
        return global_Eobj
    
    
    #==============
    def evaluate_one(self, nonant_cache,scenario_name,s):
        """ Evaluate xhat for one scenario.

        Args:
            nonant_cache(numpy vector): special numpy vector with nonant values (see ph)
            scenario_name(str): TODO
        

        Returns:
            Eobj (float or None): Expected value (or None if infeasible)

        """
        self._fix_nonants(nonant_cache)
        if not hasattr(self, "objs_dict"):
            self.objs_dict = {}
        

        solver_options = self.options["solver_options"] if "solver_options" in self.options else None
        k = scenario_name
        pyomo_solve_time = self.solve_one(solver_options,k, s,
                                          dtiming=False,
                                          verbose=self.verbose,
                                          tee=False,
                                          gripe=True,
                                          compute_val_at_nonant=True
                                          )
        
        obj = self.objs_dict[k]
        
        return obj
    
    def evaluate(self, nonant_cache, fct=None):
        """ Do the optimization and compute the expected value of the composition of the objective function 
            and a given function fct across all scenarios.

        Args:
            nonant_cache(numpy vector): special numpy vector with nonant values (see spopt)
            fct (function, optional):
                A function R-->R^p, such as x|-->(x,x^2,x^3). Default is None
                If fct is None, evaluate returns the exepected value.

        Returns:
            Eobj (float or numpy.array): Expected value

        """
        self._fix_nonants(nonant_cache)

        solver_options = self.options["solver_options"] if "solver_options" in self.options else None
        
        self.solve_loop(solver_options=solver_options,
                        use_scenarios_not_subproblems=True,
                        gripe=True, 
                        tee=False,
                        verbose=self.verbose,
                        compute_val_at_nonant=True
                        )
        
        Eobj = self.Eobjective(self.verbose,fct=fct)
        
        return Eobj
    
    #======================================================================
    def _fix_nonants_at_value(self):
        """ Fix the Vars subject to non-anticipativity at their current values.
            Loop over the scenarios to restore, but loop over subproblems
            to alert persistent solvers.
        """
        for k,s in self.local_scenarios.items():

            persistent_solver = None
            if not self.bundling:
                if (sputils.is_persistent(s._solver_plugin)):
                    persistent_solver = s._solver_plugin

            for var in s._mpisppy_data.nonant_indices.values():
                var.fix()
                if not self.bundling and persistent_solver is not None:
                    persistent_solver.update_var(var)

        if self.bundling:  # we might need to update persistent solvers
            rank_local = self.cylinder_rank
            for k,s in self.local_subproblems.items():
                if (sputils.is_persistent(s._solver_plugin)):
                    persistent_solver = s._solver_plugin
                else:
                    break  # all solvers should be the same

                # the bundle number is the last number in the name
                bunnum = sputils.extract_num(k)
                # for the scenarios in this bundle, update Vars
                for sname, scen in self.local_scenarios.items():
                    if sname not in self.names_in_bundles[rank_local][bunnum]:
                        break
                    for var in scen._mpisppy_data.nonant_indices.values():
                        persistent_solver.update_var(var)

    def calculate_incumbent(self, fix_nonants=True, verbose=False):
        """
        Calculates the current incumbent

        Args:
            solver_options (dict): passed through to the solver
            verbose (boolean): controls debugging output
        Returns:
            xhatobjective (float or None): the objective function
                or None if one could not be obtained.
        """

        if fix_nonants:
            self._fix_nonants_at_value()

        self.solve_loop(solver_options=self.current_solver_options, 
                        verbose=verbose)

        infeasP = self.infeas_prob()
        if infeasP != 0.:
            return None
        else:
            if verbose and self.cylinder_rank == 0:
                print("  Feasible xhat found")
            return self.Eobjective(verbose=verbose)
    



if __name__ == "__main__":
    #==============================
    # hardwired by dlw for debugging (this main is like MMW, but strange)
    import mpisppy.tests.examples.farmer as refmodel
    import mpisppy.utils.amalgomator as ama
    
    # do the right term of MMW (9) using the first scenarios
    ama_options = {"EF-2stage": True}   # 2stage vs. mstage
    ama_object = ama.from_module("mpisppy.tests.examples.farmer", ama_options)
    ama_object.run()
    print(f"inner bound=", ama_object.best_inner_bound)
    # This the right term of LHS of MMW (9)
    print(f"outer bound=", ama_object.best_outer_bound)

    
    ############### now get an xhat using different scenarios
    # (use use the ama args to get problem parameters)
    ScenCount = ama_object.options['num_scens']
    scenario_creator = refmodel.scenario_creator
    scenario_denouement = refmodel.scenario_denouement
    crops_multiplier = ama_object.options['crops_multiplier']
    solvername = ama_object.options['EF_solver_name']

    scenario_creator_kwargs = {
        "use_integer": False,
        "crops_multiplier": crops_multiplier,
        "num_scens": ScenCount
    }

    # different scenarios for xhat
    scenario_names = ['Scenario' + str(i) for i in range(ScenCount, 2*ScenCount)]

    ef = sputils.create_EF(
        scenario_names,
        scenario_creator,
        scenario_creator_kwargs=scenario_creator_kwargs,
    )

    solver = pyo.SolverFactory(solvername)
    if 'persistent' in solvername:
        solver.set_instance(ef, symbolic_solver_labels=True)
        solver.solve(tee=False)
    else:
        solver.solve(ef, tee=False, symbolic_solver_labels=True,)

    print(f"Xhat in-sample objective: {pyo.value(ef.EF_Obj)}")
    

    ########### get the nonants (the xhat)
    # NOTE: we probably should do an assert or two to make sure Vars match
    nonant_cache = sputils.nonant_cache_from_ef(ef)

    # Create the eval object for the left term of the LHS of (9) in MMW
    # but we are back to using the first scenarios
    MMW_scenario_names = ['scen' + str(i) for i in range(ScenCount)]

    # The options need to be re-done (and phase needs to be split up)
    options = {"iter0_solver_options": None,
             "iterk_solver_options": None,
             "solvername": solvername,
             "verbose": False}
    # TBD: set solver options
    ev = Xhat_Eval(options,
                   MMW_scenario_names,
                   scenario_creator,
                   scenario_denouement,
                   )
    obj_at_xhat = ev.evaluate(nonant_cache)
    print(f"Expected value at xhat={obj_at_xhat}")  # Left term of LHS of (9)




# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# dlw, Feb 2018. Code for closes scenario to xbar
# TODO Apr. 2020 Eliminate old references to companiondriver
import numpy as np
import mpi4py.MPI as mpi
import mpisppy.extensions.xhatbase
import pyomo.environ as pyo

class XhatClosest(mpisppy.extensions.xhatbase.XhatBase):
    """
    Args:
        rank (int): mpi process rank of currently running process
    """
    def __init__(self, ph):
        super().__init__(ph)
        self.options = ph.options["xhat_closest_options"]
        self.solver_options = self.options["xhat_solver_options"]
        self.keep_solution = True
        if ('keep_solution' in self.options) and (not self.options['keep_solution']):
            self.keep_solution = False

    def xhat_closest_to_xbar(self, verbose=False, restore_nonants=True):
        """ Get a truncated z score and look for the closest overall.

        Returns:
            obj (float or None): objective value, none if infeasible.
            snamedict (str): closest scenarios
        """
        def _vb(msg):
            if verbose and self.cylinder_rank == 0:
                print ("(rank0) xhat_looper: " + msg)

        localmindist = np.zeros(1, dtype='d')
        globalmindist = np.zeros(1, dtype='d')
        localwinnername = None
        for k, s in self.opt.local_scenarios.items():
            dist = 0
            for ndn_i, xvar in s._mpisppy_data.nonant_indices.items():
                diff = pyo.value(xvar) - pyo.value(s._mpisppy_model.xbars[ndn_i])
                variance = pyo.value(s._mpisppy_model.xsqbars[ndn_i]) \
                  - pyo.value(s._mpisppy_model.xbars[ndn_i])*pyo.value(s._mpisppy_model.xbars[ndn_i])
                if variance > 0:
                    stdev = np.sqrt(variance)
                    dist += min(3, abs(diff)/stdev)
            if localwinnername is None:
                localmindist[0] = dist
                localwinnername = k
            elif dist < localmindist[0]:
                localmindist[0] = dist
                localwinnername = k

        self.comms["ROOT"].Allreduce([localmindist, mpi.DOUBLE],
                                     [globalmindist, mpi.DOUBLE],
                                     op=mpi.MIN)
        # ties are possible, so break the tie
        localwinrank = np.zeros(1, dtype='d')  # could use a python variable.
        globalwinrank = np.zeros(1, dtype='d')
        if globalmindist[0] < localmindist[0]:
            localwinrank[0] = -1  # we lost
        else:
            localwinrank[0] = self.cylinder_rank
        self.comms["ROOT"].Allreduce([localwinrank, mpi.DOUBLE],
                                     [globalwinrank, mpi.DOUBLE],
                                     op=mpi.MAX)

        # We only used the rank to break a possible tie.
        if self.cylinder_rank == int(globalwinrank[0]):
            globalwinnername = localwinnername
        else:
            globalwinnername = None

        sroot = globalwinrank[0]

        sname = self.comms["ROOT"].bcast(globalwinnername, root=sroot)
        _vb("Trying scenario "+sname)
        _vb("   Solver options="+str(self.solver_options))
        # xxx TBD mult-stage
        snamedict = {"ROOT": sname}
        obj = self._try_one(snamedict,
                            solver_options=self.solver_options,
                            verbose=False,
                            restore_nonants=restore_nonants)
        if obj is None:
            _vb("    Infeasible")
        else:
            _vb("    Feasible, returning " + str(obj))
            
        return obj, snamedict
        
    def pre_iter0(self):
        if self.opt.multistage:
            raise RuntimeError("xhatclosest not done for multi-stage")

    def post_iter0(self):
        # a little bit silly
        self.comms = self.opt.comms
        
    def miditer(self):
        pass

    def enditer(self):
        pass

    def post_everything(self):
        # if we're keeping the solution, we *do not* restore the nonants
        restore_nonants = not self.keep_solution
        self.opt.disable_W_and_prox()
        obj, srcsname = self.xhat_closest_to_xbar(verbose=self.verbose, restore_nonants=restore_nonants)
        self.opt.reenable_W_and_prox()
        self.xhat_common_post_everything("closest to xbar", obj, srcsname, restore_nonants)
        self._final_xhat_closest_obj = obj



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# Provide some test for mpisppy.0
# Author: David L. Woodruff (circa January 2019)
"""
IMPORTANT:
  Unless we run to convergence, the solver, and even solver
version matter a lot, so we often just do smoke tests.
"""

import os
import pandas as pd
import unittest

import pyomo.environ as pyo
import mpisppy.opt.ph
import mpisppy.phbase
import mpisppy.utils.sputils as sputils
from mpisppy.tests.examples.sizes.sizes import scenario_creator, \
                                               scenario_denouement, \
                                               _rho_setter
import mpisppy.tests.examples.hydro.hydro as hydro
from mpisppy.extensions.xhatspecific import XhatSpecific
from mpisppy.tests.test_utils import get_solver,round_pos_sig

__version__ = 0.54

solver_available,solvername, persistent_available, persistentsolvername= get_solver()

def _get_ph_base_options():
    Baseoptions = {}
    Baseoptions["asynchronousPH"] = False
    Baseoptions["solvername"] = solvername
    Baseoptions["PHIterLimit"] = 10
    Baseoptions["defaultPHrho"] = 1
    Baseoptions["convthresh"] = 0.001
    Baseoptions["subsolvedirectives"] = None
    Baseoptions["verbose"] = False
    Baseoptions["display_timing"] = False
    Baseoptions["display_progress"] = False
    if "cplex" in solvername:
        Baseoptions["iter0_solver_options"] = {"mip_tolerances_mipgap": 0.001}
        Baseoptions["iterk_solver_options"] = {"mip_tolerances_mipgap": 0.00001}
    else:
        Baseoptions["iter0_solver_options"] = {"mipgap": 0.001}
        Baseoptions["iterk_solver_options"] = {"mipgap": 0.00001}

    Baseoptions["display_progress"] = False

    return Baseoptions




#*****************************************************************************
class Test_sizes(unittest.TestCase):
    """ Test the mpisppy code using sizes."""

    def setUp(self):
        self.all3_scenario_names = list()
        for sn in range(3):
            self.all3_scenario_names.append("Scenario"+str(sn+1))

        self.all10_scenario_names = list()
        for sn in range(10):
            self.all10_scenario_names.append("Scenario"+str(sn+1))

    def _copy_of_base_options(self):
        return _get_ph_base_options()

    def _fix_creator(self, scenario_name, scenario_count=None):
        """ For the test of fixed vars"""
        model = scenario_creator(
            scenario_name, scenario_count=scenario_count
        )
        model.NumProducedFirstStage[5].fix(1134) # 3scen -> 45k
        return model
        
    def test_disable_tictoc(self):
        import mpisppy.utils.sputils as utils
        print("disabling tictoc output")
        utils.disable_tictoc_output()
        # now just do anything that would make a little tictoc output
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 0

        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
        print("reeanabling tictoc ouput")
        utils.reenable_tictoc_output()


    def test_ph_constructor(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 0

        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )

    def test_ef_constructor(self):
        ScenCount = 3
        ef = mpisppy.utils.sputils.create_EF(
            self.all3_scenario_names,
            scenario_creator,
            scenario_creator_kwargs={"scenario_count": ScenCount},
            suppress_warnings=True
        )

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ef_solve(self):
        options = self._copy_of_base_options()
        solver = pyo.SolverFactory(options["solvername"])
        ScenCount = 3
        ef = mpisppy.utils.sputils.create_EF(
            self.all3_scenario_names,
            scenario_creator,
            scenario_creator_kwargs={"scenario_count": ScenCount},
            suppress_warnings=True
        )
        if '_persistent' in options["solvername"]:
            solver.set_instance(ef)
        results = solver.solve(ef, tee=False)
        sig2eobj = round_pos_sig(pyo.value(ef.EF_Obj),2)
        self.assertEqual(220000.0, sig2eobj)

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_fix_ef_solve(self):
        options = self._copy_of_base_options()
        solver = pyo.SolverFactory(options["solvername"])
        ScenCount = 3
        ef = mpisppy.utils.sputils.create_EF(
            self.all3_scenario_names,
            self._fix_creator,
            scenario_creator_kwargs={"scenario_count": ScenCount},
        )
        if '_persistent' in options["solvername"]:
            solver.set_instance(ef)
        results = solver.solve(ef, tee=False)
        sig2eobj = round_pos_sig(pyo.value(ef.EF_Obj),2)
        # the fix creator fixed num prod first stage for size 5 to 1134
        self.assertEqual(pyo.value(ef.Scenario1.NumProducedFirstStage[5]), 1134)

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ph_iter0(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 0

        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
    
        conv, obj, tbound = ph.ph_main()

        sig2obj = round_pos_sig(obj,2)
        #self.assertEqual(1400000000, sig2obj)

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_fix_ph_iter0(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 0

        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            self._fix_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
    
        conv, obj, tbound = ph.ph_main()

        # Actually, if it is still fixed for one scenario, probably fixed forall.
        for k,s in ph.local_scenarios.items():
            self.assertEqual(pyo.value(s.NumProducedFirstStage[5]), 1134)
            self.assertTrue(s.NumProducedFirstStage[5].is_fixed())
        
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ph_basic(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
        conv, obj, tbound = ph.ph_main()


    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_fix_ph_basic(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            self._fix_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
        conv, obj, tbound = ph.ph_main()
        for k,s in ph.local_scenarios.items():
            self.assertTrue(s.NumProducedFirstStage[5].is_fixed())
            self.assertEqual(pyo.value(s.NumProducedFirstStage[5]), 1134)
        

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ph_rhosetter(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
            rho_setter=_rho_setter,
        )
        conv, obj, tbound = ph.ph_main()
        sig2obj = round_pos_sig(obj,2)
        #self.assertEqual(220000.0, sig2obj)
        
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_bundles(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["bundles_per_rank"] = 2
        ph = mpisppy.opt.ph.PH(
            options,
            self.all10_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 10},
        )
        conv, obj, tbound = ph.ph_main()

    @unittest.skipIf(not persistent_available,
                     "%s solver is not available" % (persistentsolvername,))
    def test_persistent_basic(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 10
        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
        conv, basic_obj, tbound = ph.ph_main()

        options = self._copy_of_base_options()
        options["PHIterLimit"] = 10
        options["solvername"] = persistentsolvername
        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
        )
        conv, pobj, tbound = ph.ph_main()

        sig2basic = round_pos_sig(basic_obj,2)
        sig2pobj = round_pos_sig(pobj,2)
        self.assertEqual(sig2basic, sig2pobj)
        
    @unittest.skipIf(not persistent_available,
                     "%s solver is not available" % (persistentsolvername,))
    def test_persistent_bundles(self):
        """ This excercises complicated code.
        """
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["bundles_per_rank"] = 2
        ph = mpisppy.opt.ph.PH(
            options,
            self.all10_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 10},
        )
        conv, basic_obj, tbound = ph.ph_main()

        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        options["bundles_per_rank"] = 2
        options["solvername"] = persistentsolvername
        ph = mpisppy.opt.ph.PH(
            options,
            self.all10_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 10},
        )
        conv, pbobj, tbound = ph.ph_main()

        sig2basic = round_pos_sig(basic_obj,2)
        sig2pbobj = round_pos_sig(pbobj,2)
        self.assertEqual(sig2basic, sig2pbobj)

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_xhat_extension(self):
        """ Make sure least one of the xhat extensions runs.
        """
        from mpisppy.extensions.xhatlooper import XhatLooper
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 1
        options["xhat_looper_options"] =  {"xhat_solver_options":\
                                             options["iterk_solver_options"],
                                             "scen_limit": 3}

        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
            extensions=XhatLooper,
        )
        conv, basic_obj, tbound = ph.ph_main()
        # in this particular case, the extobject is an xhatter
        xhatobj1 = round_pos_sig(ph.extobject._xhat_looper_obj_final, 1)
        self.assertEqual(xhatobj1, 200000)

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_fix_xhat_extension(self):
        """ Make sure that ph and xhat do not unfix a fixed Var
        """
        from mpisppy.extensions.xhatlooper import XhatLooper
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 1
        options["xhat_looper_options"] =  {"xhat_solver_options":\
                                             options["iterk_solver_options"],
                                             "scen_limit": 3}

        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            self._fix_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
            extensions=XhatLooper,
        )
        conv, basic_obj, tbound = ph.ph_main()

        for k,s in ph.local_scenarios.items():
            self.assertTrue(s.NumProducedFirstStage[5].is_fixed())
            self.assertEqual(pyo.value(s.NumProducedFirstStage[5]), 1134)

            
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_lagrangian_bound(self):
        """ Make sure the lagrangian bound is at least a bound
        """
        from mpisppy.extensions.xhatlooper import XhatLooper
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 1
        options["xhat_looper_options"] =  {"xhat_solver_options":\
                                             options["iterk_solver_options"],
                                             "scen_limit": 3}
        ph = mpisppy.opt.ph.PH(
            options,
            self.all3_scenario_names,
            scenario_creator,
            scenario_denouement,
            scenario_creator_kwargs={"scenario_count": 3},
            extensions=XhatLooper,
        )
        conv, basic_obj, tbound = ph.ph_main()
        xhatobj = ph.extobject._xhat_looper_obj_final
        dopts = sputils.option_string_to_dict("mipgap=0.0001")
        objbound = ph.post_solve_bound(solver_options=dopts, verbose=False)
        self.assertGreaterEqual(xhatobj+1.e-6, objbound)
        
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def smoke_for_extensions(self):
        """ Make sure the example extensions can at least run.
        """
        from mpisppy.extensions.extension import MultiExtension
        from mpisppy.extensions.fixer import Fixer
        from mpisppy.extensions.mipgapper import Gapper
        from mpisppy.extensions.xhatlooper import XhatLooper
        from mpisppy.extensions.xhatclosest import XhatClosest

        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        ph = mpisppy.opt.ph.PH(options, self.all3_scenario_names,
                                    scenario_creator, scenario_denouement,
                                    extensions=MultiExtension,
                                    extension_kwargs=multi_ext,
        )
        multi_ext = {"ext_classes": [Fixer, Gapper, XhatLooper, XhatClosest]}
        conv, basic_obj, tbound = ph.ph_main()


    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def fix_for_extensions(self):
        """ Make sure the example extensions don't destroy fixedness
        """
        from mpisppy.extensions.extensions import MultiExtension
        from mpisppy.extensions.fixer import Fixer
        from mpisppy.extensions.mipgapper import Gapper
        from mpisppy.extensions.xhatlooper import XhatLooper
        from mpisppy.extensions.xhatclosest import XhatClosest
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 2
        ph = mpisppy.opt.ph.PH(options, self.all3_scenario_names,
                                    self._fix_creator, scenario_denouement,
                                    extensions=MultiExtension,
                                    extension_kwargs=multi_ext,
        )
        multi_ext = {"ext_classes": [Fixer, Gapper, XhatLooper, XhatClosest]}
        conv, basic_obj, tbound = ph.ph_main( )
        for k,s in ph.local_scenarios.items():
            self.assertTrue(s.NumProducedFirstStage[5].is_fixed())
            self.assertEqual(pyo.value(s.NumProducedFirstStage[5]), 1134)

#*****************************************************************************
class Test_hydro(unittest.TestCase):
    """ Test the mpisppy code using hydro (three stages)."""

    def _copy_of_base_options(self):
        retval = _get_ph_base_options()
        retval["branching_factors"] = [3, 3]
        return retval
    

    def setUp(self):
        self.options = self._copy_of_base_options()
        # branching factor (3 stages is hard-wired)
        self.BFs = self.options["branching_factors"]
        self.ScenCount = self.BFs[0] * self.BFs[1]
        self.all_scenario_names = list()
        for sn in range(self.ScenCount):
            self.all_scenario_names.append("Scen"+str(sn+1))
        self.all_nodenames = ["ROOT"] # all trees must have this node
        for b in range(self.BFs[0]):
            self.all_nodenames.append("ROOT_"+str(b))
        # end hardwire

    def test_ph_constructor(self):
        options = self._copy_of_base_options()

        ph = mpisppy.opt.ph.PH(
            options,
            self.all_scenario_names,
            hydro.scenario_creator,
            hydro.scenario_denouement,
            all_nodenames=self.all_nodenames,
            scenario_creator_kwargs={"branching_factors": self.BFs},
        )

    def test_ef_constructor(self):
        ef = mpisppy.utils.sputils.create_EF(
            self.all_scenario_names,
            hydro.scenario_creator,
            scenario_creator_kwargs={"branching_factors": self.BFs},
        )

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ef_solve(self):
        options = self._copy_of_base_options()
        solver = pyo.SolverFactory(options["solvername"])
        ef = mpisppy.utils.sputils.create_EF(
            self.all_scenario_names,
            hydro.scenario_creator,
            scenario_creator_kwargs={"branching_factors": self.BFs},
        )
        if '_persistent' in options["solvername"]:
            solver.set_instance(ef)
        results = solver.solve(ef, tee=False)
        mpisppy.utils.sputils.ef_nonants_csv(ef, "delme.csv")

        df = pd.read_csv("delme.csv", index_col=1)
        val2 = round_pos_sig(df.loc[" Scen7.Pgt[2]"].at[" Value"])
        self.assertEqual(val2, 60)
        os.remove("delme.csv")

        
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ef_csv(self):
        options = self._copy_of_base_options()
        solver = pyo.SolverFactory(options["solvername"])
        ef = mpisppy.utils.sputils.create_EF(
            self.all_scenario_names,
            hydro.scenario_creator,
            scenario_creator_kwargs={"branching_factors": self.BFs},
        )
        if '_persistent' in options["solvername"]:
            solver.set_instance(ef)
        results = solver.solve(ef, tee=False)
    

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ph_solve(self):
        options = self._copy_of_base_options()

        ph = mpisppy.opt.ph.PH(
            options,
            self.all_scenario_names,
            hydro.scenario_creator,
            hydro.scenario_denouement,
            all_nodenames=self.all_nodenames,
            scenario_creator_kwargs={"branching_factors": self.BFs},
        )
        conv, obj, tbound = ph.ph_main()

        sig2tbnd = round_pos_sig(tbound, 2)
        self.assertEqual(180, sig2tbnd)

        ph.disable_W_and_prox()
        sig2eobj = round_pos_sig(ph.Eobjective(), 2)
        self.assertEqual(190, sig2eobj)

        
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ph_xhat(self):
        options = self._copy_of_base_options()
        options["PHIterLimit"] = 10  # xhat is feasible
        options["xhat_specific_options"] = {"xhat_solver_options":
                                              options["iterk_solver_options"],
                                              "xhat_scenario_dict": \
                                              {"ROOT": "Scen1",
                                               "ROOT_0": "Scen1",
                                               "ROOT_1": "Scen4",
                                               "ROOT_2": "Scen7"},
                                              "csvname": "specific.csv"}

        ph = mpisppy.opt.ph.PH(
            options,
            self.all_scenario_names,
            hydro.scenario_creator,
            hydro.scenario_denouement,
            all_nodenames=self.all_nodenames,
            scenario_creator_kwargs={"branching_factors": self.BFs},
            extensions = XhatSpecific
        )
        conv, obj, tbound = ph.ph_main()
        sig2xhatobj = round_pos_sig(ph.extobject._xhat_specific_obj_final, 2)
        self.assertEqual(190, sig2xhatobj)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# Try a particular scenario as xhat. This is mainly for regression testing.
# DLW, Feb 2019
# This extension uses options["xhat_scenario_dict"] (keys are node names)

import mpisppy.extensions.xhatbase

class XhatSpecific(mpisppy.extensions.xhatbase.XhatBase):
    """
    Args:
        spo (SPOpt object): the calling object
        rank (int): mpi process rank of currently running process
    """
    def __init__(self, spo):
        super().__init__(spo)
        self.options = spo.options["xhat_specific_options"]
        self.solver_options = self.options["xhat_solver_options"]
        self.keep_solution = True
        if ('keep_solution' in self.options) and (not self.options['keep_solution']):
            self.keep_solution = False

    #==========
    def xhat_tryit(self,
                   xhat_scenario_dict,
                   verbose=False,
                   restore_nonants=True):
        """If your rank has
        the chosen guy, bcast, if not, recieve the bcast. In any event, fix the vars
        at the bcast values and see if it is feasible. 

        Args:
            xhat_scenario_dict (string): keys are nodes; values are scen names
            verbose (boolean): controls debugging output
        Returns:
            xhatobjective (float or None): the objective function
                or None if one could not be obtained.
        """
        def _vb(msg):
            if verbose and self.cylinder_rank == 0:
                print("  xhat_specific: " + msg)

        obj = None
        sname = None

        _vb("Enter XhatSpecific.xhat_tryit to try: "+str(xhat_scenario_dict))

        _vb("   Solver options="+str(self.solver_options))
        obj = self._try_one(xhat_scenario_dict,
                            solver_options=self.solver_options,
                            verbose=False,
                            restore_nonants=restore_nonants)
        if obj is None:
            _vb("Infeasible")
        else:
            _vb("Feasible, returning " + str(obj))

        return obj

    def pre_iter0(self):
        pass

    def post_iter0(self):
        # a little bit silly
        self.comms = self.opt.comms
        
    def miditer(self):
        pass

    def enditer(self):
        pass

    def post_everything(self):
        # if we're keeping the solution, we *do not* restore the nonants
        restore_nonants = not self.keep_solution
        self.opt.disable_W_and_prox()
        xhat_scenario_dict = self.options["xhat_scenario_dict"]
        obj = self.xhat_tryit(xhat_scenario_dict,
                              verbose=self.verbose,
                              restore_nonants=restore_nonants)
        self.opt.reenable_W_and_prox()
        # to make available to tester
        self._xhat_specific_obj_final = obj
        self.xhat_common_post_everything("xhat specified scenario", obj, xhat_scenario_dict, restore_nonants)



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
""" Code for WW style fixing of integers. This can be used
    as the only extension, but it could be called from a "multi"
    extension.
"""
"""
Updated Feb 2019.
The options give lists of (id, sqrt_thresh, NB_lag, LB_lag, UB_lag) 
If the id is not there, don't fix (5-tuple).
There is one for iter0 and one for iterk.
For iter0, if the lag is not None, do it.
For other iters, use count;  None is also how you avoid.
"""

import pyomo.environ as pyo
import mpisppy.utils.sputils as sputils
import mpisppy.extensions.extension

def Fixer_tuple(xvar, th=None, nb=None, lb=None, ub=None):
    """ Somewhat self-documenting way to make a fixer tuple.
        For use in/by/for the so-called Reference Model.
    Args:
        xvar (Var): the Pyomo Var
        th (float): compared to sqrt(abs(xbar_sqared - xsquared_bar))
        nb: (int) None means ignore; for iter k, number of iters 
                  converged anywhere.
        lb: (int) None means ignore; for iter k, number of iters 
                  converged within th of xvar lower bound.
        ub: (int) None means ignore; for iter k, number of iters
                  converged within th of xvar upper bound

    Returns:
        tuple: a tuple to be appended to the iter0 or iterk list of tuples
    """
    if th is None and nb is None and lb is None and ub is None:
        print ("warning: Fixer_tuple called for Var=", xvar.name,
               "but no arguments were given")
    if th is None:
        th = 0
    if nb is not None and lb is not None and nb < lb:
        print ("warning: Fixer_tuple called for Var=", xvar.name,
               "with nb < lb, which means lb will be ignored.")
    if  nb is not None and ub is not None and nb < ub:
        print ("warning: Fixer_tuple called for Var=", xvar.name,
               "with nb < ub, which means ub will be ignored.")
        
    return (id(xvar), th, nb, lb, ub)

class Fixer(mpisppy.extensions.extension.Extension):

    def __init__(self, ph):
        self.ph = ph
        self.cylinder_rank = self.ph.cylinder_rank
        self.options = ph.options
        self.fixeroptions = self.options["fixeroptions"] # required
        self.verbose = self.options["verbose"] \
                       or self.fixeroptions["verbose"]
        # this function is scenario specific (takes a scenario as an arg)
        self.id_fix_list_fct = self.fixeroptions["id_fix_list_fct"]
        self.dprogress = ph.options["display_progress"]
        self.fixed_so_far = 0
        self.boundtol = self.fixeroptions["boundtol"]

    def populate(self, local_scenarios):
        # [(ndn, i)] = iter count (i indices into nonantlist)
        self.local_scenarios = local_scenarios

        self.iter0_fixer_tuples = {} # caller data
        self.fixer_tuples = {} # caller data
        self.threshold = {}  
        self.iter0_threshold = {}  
        # This count dict drives the loops later
        # NOTE: every scenario has a list.
        for k,s in self.local_scenarios.items():
            s._mpisppy_data.conv_iter_count = {} # key is (ndn, i)

            self.iter0_fixer_tuples[s], self.fixer_tuples[s] = \
                    self.id_fix_list_fct(s)

            if self.iter0_fixer_tuples[s] is not None:
                for (varid, th, nb, lb, ub) in self.iter0_fixer_tuples[s]:
                    (ndn, i) = s._mpisppy_data.varid_to_nonant_index[varid] #
                    if (ndn, i) not in self.iter0_threshold:
                        self.iter0_threshold[(ndn, i)] = th
                    else:
                        if th != self.iter0_threshold[(ndn, i)]:
                            print (s.name, ndn, i, th)
                            raise RuntimeError("Attempt to vary iter0 fixer "+\
                                               "threshold across scenarios.")
            if self.fixer_tuples[s] is not None:
                for (varid, th, nb, lb, ub) in self.fixer_tuples[s]:
                    (ndn, i) = s._mpisppy_data.varid_to_nonant_index[varid]
                    s._mpisppy_data.conv_iter_count[(ndn, i)] = 0
                    if (ndn, i) not in self.threshold:
                        self.threshold[(ndn, i)] = th
                    else:
                        if th != self.threshold[(ndn, i)]:
                            print (s.name, ndn, i, th)
                            raise RuntimeError("Attempt to vary fixer "+\
                                               "threshold across scenarios")

    def _vb(self, str):
        if self.verbose and self.cylinder_rank == 0:
            print ("(rank0) " + str)

    def _dp(self, str):
        if (self.dprogress or self.verbose) and self.cylinder_rank == 0:
            print ("(rank0) " + str)

    def _update_fix_counts(self):
        nodesdone = []  # avoid multiple updates of a node's Vars
        for k,s in self.local_scenarios.items():
            for ndn_i, xvar in s._mpisppy_data.nonant_indices.items():
                if xvar.is_fixed():
                    continue
                xb = pyo.value(s._mpisppy_model.xbars[ndn_i])
                diff = xb * xb - pyo.value(s._mpisppy_model.xsqbars[ndn_i])
                tolval = self.threshold[ndn_i]
                tolval *= tolval  # the tol is on sqrt
                if -diff < tolval and diff < tolval:
                    ##print ("debug += diff, tolval", diff, tolval)
                    s._mpisppy_data.conv_iter_count[ndn_i] += 1
                else:
                    s._mpisppy_data.conv_iter_count[ndn_i] = 0
                    ##print ("debug reset fix diff, tolval", diff, tolval)
                    
    def iter0(self, local_scenarios):

        # first, do some persistent solver with bundles gymnastics        
        have_bundles = hasattr(self.ph, "saved_objs") # indicates bundles
        if have_bundles:
            subpname = next(iter(self.ph.local_subproblems))
            subp = self.ph.local_subproblems[subpname]
            solver_is_persistent = isinstance(subp._solver_plugin,
            pyo.pyomo.solvers.plugins.solvers.persistent_solver.PersistentSolver)
            if solver_is_persistent:
                vars_to_update = {}

        fixoptions = self.fixeroptions
        raw_fixed_so_far = 0   # count those fixed in each scenario
        for sname,s in self.ph.local_scenarios.items():
            if self.iter0_fixer_tuples[s] is None:
                print ("WARNING: No Iter0 fixer tuple for s.name=",s.name)
                return
            if not have_bundles:
                solver_is_persistent = isinstance(s._solver_plugin,
                    pyo.pyomo.solvers.plugins.solvers.persistent_solver.PersistentSolver)

            for (varid, th, nb, lb, ub) in self.iter0_fixer_tuples[s]:
                was_fixed = False
                try:
                    (ndn, i) = s._mpisppy_data.varid_to_nonant_index[varid]
                except:
                    print ("Are you trying to fix a Var that is not nonant?")
                    raise
                xvar = s._mpisppy_data.nonant_indices[ndn,i]
                if not xvar.is_fixed():
                    xb = pyo.value(s._mpisppy_model.xbars[(ndn,i)])
                    diff = xb * xb - pyo.value(s._mpisppy_model.xsqbars[(ndn,i)])
                    tolval = self.iter0_threshold[(ndn, i)]
                    sqtolval = tolval*tolval  # the tol is on sqrt
                    if -diff > sqtolval or diff > sqtolval:
                        ##print ("debug0 NO fix diff, sqtolval", diff, sqtolval)
                        continue
                    else:
                        ##print ("debug0 fix diff, sqtolval", diff, sqtolval)
                        # if we are still here, it is converged
                        if nb is not None:
                            xvar.fix(xb)
                            self._vb("Fixed0 nb %s %s at %s" % \
                                     (s.name, xvar.name, str(xvar._value)))
                            was_fixed = True
                        elif lb is not None and xb - xvar.lb < self.boundtol:
                            xvar.fix(xvar.lb)
                            self._vb("Fixed0 lb %s %s at %s" % \
                                     (s.name, xvar.name, str(xvar._value)))
                            was_fixed = True
                        elif ub is not None and xvar.ub - xb < self.boundtol:
                            xvar.fix(xvar.ub)
                            self._vb("Fixed0 ub %s %s at %s" % \
                                     (s.name, xvar.name, str(xvar._value)))
                            was_fixed = True

                    if was_fixed:
                        raw_fixed_so_far += 1
                        if not have_bundles and solver_is_persistent:
                            s._solver_plugin.update_var(xvar)
                        if have_bundles and solver_is_persistent:
                            if sname not in vars_to_update:
                                vars_to_update[sname] = []
                            vars_to_update[sname].append(xvar)

        if have_bundles and solver_is_persistent:
            for k,subp in self.ph.local_subproblems.items():
                subpnum = sputils.extract_num(k)
                rank_local = self.ph.cylinder_rank
                for sname in self.ph.names_in_bundles[rank_local][subpnum]:
                    if sname in vars_to_update:
                        for xvar in vars_to_update[sname]:
                            subp._solver_plugin.update_var(xvar)
                        
        self.fixed_so_far += raw_fixed_so_far / len(local_scenarios)
        self._dp("Unique Vars fixed so far %s" % (self.fixed_so_far))
        if raw_fixed_so_far % len(local_scenarios) != 0:
            raise RuntimeError ("Variation in fixing across scenarios detected "
                                "in fixer.py (iter0)")
            # maybe to do mpicomm.abort()

    def iterk(self, PHIter):
        """ Before iter k>1 solves, but after x-bar update.
        """
        # first, do some persistent solver with bundles gymnastics        
        have_bundles = hasattr(self.ph, "saved_objs") # indicates bundles
        if have_bundles:
            subpname = next(iter(self.ph.local_subproblems))
            subp = self.ph.local_subproblems[subpname]
            solver_is_persistent = isinstance(subp._solver_plugin,
            pyo.pyomo.solvers.plugins.solvers.persistent_solver.PersistentSolver)
            if solver_is_persistent:
                vars_to_update = {}

        fixoptions = self.fixeroptions
        raw_fixed_so_far = 0
        self._update_fix_counts()
        for sname,s in self.local_scenarios.items():
            if self.fixer_tuples[s] is None:
                print ("MAJOR WARNING: No Iter k fixer tuple for s.name=",s.name)
                return
            if not have_bundles:
                solver_is_persistent = isinstance(s._solver_plugin, pyo.pyomo.solvers.plugins.solvers.persistent_solver.PersistentSolver)
            for (varid, th, nb, lb, ub) in self.fixer_tuples[s]:
                was_fixed = False
                try:
                    (ndn, i) = s._mpisppy_data.varid_to_nonant_index[varid]
                except:
                    print ("Are you trying to fix a Var that is not nonant?")
                    raise
                tolval = self.threshold[(ndn, i)]
                xvar = s._mpisppy_data.nonant_indices[ndn,i]
                if not xvar.is_fixed():
                    xb = pyo.value(s._mpisppy_model.xbars[(ndn,i)])
                    fx = s._mpisppy_data.conv_iter_count[(ndn,i)]
                    if fx > 0:
                        xbar = pyo.value(s._mpisppy_model.xbars[(ndn,i)])
                        was_fixed = False
                        if  nb is not None and nb <= fx:
                            xvar.fix(xbar)
                            self._vb("Fixed nb %s %s at %s" % \
                                     (s.name, xvar.name, str(xvar._value)))
                            was_fixed = True
                        elif lb is not None and lb < fx \
                             and xb - xvar.lb < self.boundtol:
                            xvar.fix(xvar.lb)
                            self._vb("Fixed lb %s %s at %s" % \
                                     (s.name, xvar.name, str(xvar._value)))
                            was_fixed = True
                        elif ub is not None and ub < fx \
                             and xvar.ub - xb < self.boundtol:
                            xvar.fix(xvar.ub)
                            self._vb("Fixed ub %s %s at %s" % \
                                     (s.name, xvar.name, str(xvar._value)))
                            was_fixed = True

                    if was_fixed:
                        raw_fixed_so_far += 1
                        if not have_bundles and solver_is_persistent:
                            s._solver_plugin.update_var(xvar)
                        if have_bundles and solver_is_persistent:
                            if sname not in vars_to_update:
                                vars_to_update[sname] = []
                            vars_to_update[sname].append(xvar)


        if have_bundles and solver_is_persistent:
            for k,subp in self.ph.local_subproblems.items():
                subpnum = sputils.extract_num(k)
                rank_local = self.ph.cylinder_rank
                for sname in self.ph.names_in_bundles[rank_local][subpnum]:
                    if sname in vars_to_update:
                        for xvar in vars_to_update[sname]:
                            subp._solver_plugin.update_var(xvar)
                        
        self.fixed_so_far += raw_fixed_so_far / len(self.local_scenarios)
        self._dp("Unique Vars fixed so far %s" % (self.fixed_so_far))
        if raw_fixed_so_far % len(self.local_scenarios) != 0:
            raise RuntimeError ("Variation in fixing across scenarios detected "
                                "in fixer.py")

    def pre_iter0(self):
        return
                                        
    def post_iter0(self):
        """ initialize data structures; that's all we can do at this point
        """
        self.populate(self.ph.local_scenarios)

    def miditer(self):
        """ Check for fixing before in the middle of PHIter (after
        the xbar update for PHiter-1).
        """
        PHIter = self.ph._PHIter
        if PHIter == 1:  # before iter 1 solves
            self.iter0(self.ph.local_scenarios)
        else:
            self.iterk(PHIter)

    def enditer(self):
        return

    def post_everything(self):
        self._dp("Final unique Vars fixed by fixer= %s" % \
                      (self.fixed_so_far))

